# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Quantized model calibration tools"""

import keras
import tensorflow as tf

from ..quantization_layers import (QuantizedReLU, QuantizedDense,
                                   QuantizedConv2D, QuantizedSeparableConv2D)
from ..quantization_ops import WeightQuantizer

from ..transforms.clone import clone_model_with_weights


def create_submodel(model, first_layer, last_layer):
    """Create a submodel of a Keras Model

    Args:
        model (:obj:`keras.Model`): a Keras Model
        first_layer (:obj:`keras.Layer`): the submodel first layer
        last_layer (:obj:`keras.Layer`): the submodel last layer

    Returns:
        :obj:`keras.Model`: a Keras sub-model
    """
    submodel = keras.Sequential()
    input_shape = first_layer.input_shape[1:]
    submodel.add(keras.layers.Input(shape=input_shape))
    add_layer = False
    for layer in model.layers:
        if layer == first_layer:
            add_layer = True
        if add_layer:
            submodel.add(layer)
        if layer == last_layer:
            break
    return submodel


class NoneQuantizer(WeightQuantizer):
    """A WeightQuantizer that returns float weights
    """

    def scale_factor(self, _):
        return 1

    def quantize(self, w):
        return w


def disable_quantizers(layer):
    """Disable quantizers for the specified layer

    Args:
        layer (:obj:`keras.Layer`): the layer

    Returns:
        [:obj:`cnn2snn.quantization_ops.WeightQuantizer`]: the list of disabled
        quantizers.
    """
    quantizers = [layer.quantizer]
    layer.quantizer = NoneQuantizer(layer.quantizer.bitwidth)
    if hasattr(layer, 'quantizer_dw'):
        quantizers.append(layer.quantizer_dw)
        layer.quantizer_dw = NoneQuantizer(layer.quantizer_dw.bitwidth)
    return quantizers


def restore_quantizers(layer, quantizers):
    """Restores quantizers for the specified layer

    Args:
        layer (:obj:`keras.Layer`): the layer
        quantizers ([:obj:`cnn2snn.quantization_ops.WeightQuantizer`]): the list of disabled
            quantizers.
    """
    layer.quantizer = quantizers[0]
    if hasattr(layer, 'quantizer_dw'):
        layer.quantizer_dw = quantizers[1]


def next_activation(layer):
    """Find the next activation after the specified layer

    Args:
        layer (:obj:`keras.Layer`): the layer

    Returns:
        :obj:`keras.Layer`: the next activation layer or None

    """
    outbound_nodes = layer.outbound_nodes
    for node in outbound_nodes:
        next_layer = node.layer
        if isinstance(next_layer, QuantizedReLU):
            return next_layer
        activation = next_activation(next_layer)
        if activation is not None:
            return activation
    return None


def is_on_top(layer, bottom_layer):
    """Checks if a layer follows another layer

    Args:
        layer (:obj:`keras.Layer`): the top layer
        bottom_layer (:obj:`keras.Layer`): the previous layer

    Returns:
        bool: True if the layer is on top
    """
    outbound_nodes = bottom_layer.outbound_nodes
    if len(outbound_nodes) == 0:
        # Bottom layer has no successors
        return False
    for node in outbound_nodes:
        # Check if layer is on top of one of the outbound layers
        next_layer = node.layer
        if next_layer == layer or is_on_top(layer, next_layer):
            return True
    return False


def is_quantized_neural(layer):
    """Check is a layer is a quantized neural layer

    These layers are the only ones eligible for calibration.

    Args:
        layer (:obj:`keras.Layer`): the top layer

    Returns:
        bool: True if the layer can be calibrated
    """
    return isinstance(
        layer, (QuantizedDense, QuantizedConv2D, QuantizedSeparableConv2D))


class QuantizationSampler():
    """A tool to inspect the layer outputs of a quantized model

    The sampler is initialized with a quantized model and a set of samples used
    for the evaluation of the layer outputs. An optional batch size can be
    specified to avoid out-of-memory errors when using a GPU.

    To evaluate the outputs of a specific layer, it must first be selected
    using the `select_layer` member.

    Once done, three methods are available to inspect the layer outputs:

    - `quantized_outputs` returns the actual outputs of the layer,
    - `float_outputs` returns the outputs of the layer if its weights were not
      quantized,
    - `quantization_error` applies the `keras.metrics.Metric` passed as
      arguments to the difference between the float and quantized outputs.

    Example:

        >>> # Evaluate the quantization MSE of a quantized layer
        >>> model = tf.keras.Sequential([
        ...     tf.keras.layers.Dense(5, input_shape=(3,)),
        ...     tf.keras.layers.ReLU()])
        >>> model_quantized = cnn2snn.quantize(model,
        ...                                    weight_quantization=4,
        ...                                    activ_quantization=4)
        >>> # Instantiate a QuantizationSampler with a few dataset samples
        >>> sampler = QuantizationSampler(model_quantized, samples)
        >>> # Select the quantized layer
        >>> sampler.select_layer(model_quantized.layers[0])
        >>> # Evaluate the Mean Squared Error
        >>> m = keras.metrics.MeanSquaredError()
        >>> mse = sampler.quantization_error(m)

    Args:
        model (:obj:`keras.Model`): a quantized Keras model
        samples (:obj:`np.ndarray`): a set of calibration samples
        batch_size (int): the batch size used for evaluation.
    """

    def __init__(self, model, samples, batch_size=None):
        self.model = model
        n_inputs = samples.shape[0]
        if batch_size is None:
            batch_size = n_inputs
        # We only process full batches
        self.n_batches = n_inputs // batch_size
        # Split samples by batches
        self.samples = []
        for i in range(self.n_batches):
            b_from = i * batch_size
            b_to = (i + 1) * batch_size
            self.samples.append(samples[b_from:b_to, :, :, :])
        self.inputs = None
        self.submodel = None
        self.layer = None

    def select_layer(self, layer, include_activation=False):
        """Select the layer to inspect

        Args:
            layer (:obj:`keras.Layer`): a quantized keras layer
            include_activation (bool): evaluate the outputs after activation
        """
        if not is_quantized_neural(layer):
            raise ValueError("Output layer must be a Quantized layer")
        # Evaluate the outputs immediately before the selected layer to save
        # computation time when repeatedly evaluating the layer outputs
        if self.submodel is None or not is_on_top(layer, self.layer):
            # Evaluate layer inputs from the beginning of the model
            input_layer = self.model.layers[0]
            self.inputs = []
            for i in range(self.n_batches):
                self.inputs.append(self.samples[i])
        else:
            # Evaluate layer inputs using the outputs of the current submodel
            input_layer = self.submodel.layers[-1].outbound_nodes[0].layer
            for i in range(self.n_batches):
                self.inputs[i] = self.quantized_outputs(i)
        if layer != input_layer:
            # Advance inputs
            prev_layer = layer.inbound_nodes[0].inbound_layers
            submodel = create_submodel(self.model, input_layer, prev_layer)
            for i in range(self.n_batches):
                self.inputs[i] = submodel(self.inputs[i])
        # Create the submodel including the layer (and optionally its activation)
        activation = next_activation(layer) if include_activation else None
        if activation is not None:
            self.submodel = create_submodel(self.model, layer, activation)
        else:
            self.submodel = create_submodel(self.model, layer, layer)
        # Remember current layer
        self.layer = layer

    def quantized_outputs(self, batch_index=0):
        """Evaluates the quantized outputs of the selected layer

        Args:
            batch_index (int): the samples batch index

        Returns:
            :obj:`tf.Tensor`: the quantized outputs
        """
        if self.submodel is None:
            raise SystemError("Set output layer before requesting data")
        if batch_index < 0 or batch_index >= self.n_batches:
            raise ValueError("Invalid batch index")
        return self.submodel(self.inputs[batch_index])

    def float_outputs(self, batch_index=0):
        """Evaluates the 'float' outputs of the selected layer

        This disables the weights quantization and evaluates the outputs of the
        layer as if it was using float weights.

        Args:
            batch_index (int): the samples batch index

        Returns:
            :obj:`tf.Tensor`: the quantized outputs
        """
        if self.submodel is None:
            raise SystemError("Set output layer before requesting data")
        if batch_index < 0 or batch_index >= self.n_batches:
            raise ValueError("Invalid batch index")
        quantizers = disable_quantizers(self.layer)
        outputs = self.submodel(self.inputs[batch_index])
        restore_quantizers(self.layer, quantizers)
        return outputs

    def quantization_error(self, metric):
        """Evaluates the quantization error of the selected layer

        This evaluates the difference between the layer quantized outputs and
        the outputs if the layer was using float weights using the Metric
        passed as parameter.

        Args:
            metric (:obj:`keras.metrics.Metric`): the Metric for evaluation

        Returns:
            the Metric results (usually a single float)

        """
        metric.reset_state()
        for i in range(self.n_batches):
            metric.update_state(self.float_outputs(i),
                                self.quantized_outputs(i))
        return metric.result().numpy()


def previous_neural(layer):
    """Find the previous neural layer before the specified layer

    Args:
        layer (:obj:`keras.Layer`): the layer

    Returns:
        :obj:`keras.Layer`: the next activation layer or None

    """
    inbound_nodes = layer.inbound_nodes
    for node in inbound_nodes:
        previous_layer = node.inbound_layers
        if is_quantized_neural(previous_layer):
            return previous_layer
        neural_layer = previous_neural(previous_layer)
        if neural_layer is not None:
            return neural_layer
    return None


def activations_rescaling(model, samples, batch_size=None):
    """iterate over the layers of a quantized model, evaluating
    the pre-activations before each QuantizedReLu layer,
    and adjusting the max_value parameter of each QuantizedReLu
    accordingly.

    Args:
        model (:obj:`keras.Model`): a Keras Model
        samples (:obj:`np.ndarray`): a set of calibration samples
        batch_size (int): the batch size used for evaluation.

    """
    clone_model = clone_model_with_weights(model)
    sampler = QuantizationSampler(clone_model, samples, batch_size)
    for layer in clone_model.layers:
        if not isinstance(layer, QuantizedReLU):
            continue
        previous_neural_layer = previous_neural(layer)
        sampler.select_layer(previous_neural_layer, True)
        float_output = sampler.float_outputs()
        max_val = tf.reduce_max(float_output)
        if max_val < layer.max_value_:
            layer.max_value_ = tf.constant(max_val, dtype=tf.float32)

    return clone_model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedShiftmax to Akida.
"""
from akida import Shiftmax
from quantizeml.layers import QuantizedShiftmax
import numpy as np

from ..akida_versions import AkidaVersion
from .layer_utils import get_inbound_layers
from .weights import broadcast_and_set_variable
from .block_converter import BlockConverter, register_conversion_patterns


__all__ = ["ShiftmaxBlockConverter"]

_PATTERNS = [(QuantizedShiftmax,)]


def _set_shiftmax_variables(ak_layer, block):
    """Computes and sets the variables for an Akida Shiftmax layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        ak_layer (:obj:`akida.Layer`): the targeted akida layer.
        block (list(:obj:`tf.keras.Layer`)): the block of keras layers.
    """
    input_shift = block[0].input_shift.value.numpy().astype(
        np.uint8)
    broadcast_and_set_variable(ak_layer.variables, "input_shift", input_shift)


def _create_shiftmax(block):
    """Parses a quantizeml QuantizedShiftmax layer and returns the
    corresponding Akida Shiftmax layer.

    Args:
        block (list(:obj:`tf.keras.Layer`)): list of quantizeml quantized layers.

    Returns:
        :obj:`akida.Shiftmax`: The created akida layer.
    """
    shiftmax = block[0]
    return Shiftmax(output_bits=shiftmax.exp_bitwidth + 1,
                    buffer_bits=shiftmax.buffer_bitwidth,
                    name=shiftmax.name)


def convert_quantized_shiftmax(model_ak, block):
    """Converts QuantizedShiftmax layer and its variables and adds it to the
    Akida's v2 model.

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the model will be added.
        block (list(:obj:`tf.keras.Layer`)): list of quantizeml quantized layers.

    Returns:
        bool: Returns True for a successful conversion.
    """
    # Retrieve the akida inbound layers
    inbound_layers_ak = get_inbound_layers(model_ak, block[0])
    # Create and add layer to the akida model
    layer_ak = _create_shiftmax(block)
    model_ak.add(layer_ak, inbound_layers_ak)
    # Set the akida layer converted variables
    _set_shiftmax_variables(layer_ak, block)
    return True


class ShiftmaxBlockConverter(BlockConverter):
    """Main class that should be used to check if the Shiftmax layer block is compatible to an
    Akida v2 conversion and provides a method to convert it in an equivalent Akida Shiftmax layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def convert(self, model_ak):
        return convert_quantized_shiftmax(model_ak, self._block)


# Register the valid shiftmax block pattern for Akida v2
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, ShiftmaxBlockConverter)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["DequantizerOnnxConverter"]

import akida

from .base_converter import OnnxConverter
from .weights import broadcast_and_set_variable
from .register import register_onnx_converter_target


@register_onnx_converter_target("Dequantizer")
class DequantizerOnnxConverter(OnnxConverter):
    """Convert Dequantizer node into an akida.Dequantizer.

    Args:
        node (NodeProto): the node to convert.
        model (ModelProto): the model that the node is.
    """

    def _parse_akida_layer(self):
        return akida.Dequantizer(name=self.name)

    def _set_akida_variables(self, ak_layer):
        broadcast_and_set_variable(ak_layer.variables, "scales", self.weights["scale"])



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Utility function to convert layer to akida.
"""

from quantizeml import layers as qlayers


def get_inbound_layers(ak_model, k_layer):
    """Returns a list of Akida layers in an Akida model that correspond to the inbound
    layers of a Keras layer. The list will be built by traversing the Keras model graph.

    Args:
        ak_model (:obj:`Model`): the model where to find the inbounds layers.
        k_layer (:obj:`tf.keras.Layer`): the source layer.

    Returns:
        :list: the inbounds layers names.
    """

    # The list of supported Keras layers not represented in Akida
    skippable_layers = (qlayers.QuantizedDropout, qlayers.QuantizedFlatten,
                        qlayers.QuantizedReshape, qlayers.QuantizedRescaling,
                        qlayers.QuantizedReLU, qlayers.QuantizedMaxPool2D,
                        qlayers.QuantizedGlobalAveragePooling2D,
                        qlayers.QuantizedAddPositionEmbs,
                        qlayers.QuantizedClassToken)

    # Get inbound layers names
    in_node = k_layer.inbound_nodes[0]
    inbound_layers = in_node.inbound_layers
    if not isinstance(inbound_layers, list):
        inbound_layers = [inbound_layers]
    non_skippable_inbounds = []
    for layer in inbound_layers:
        if type(layer) in skippable_layers:
            # All skippable layers have one inbound layer, so this can be
            # easily resolved.
            while type(layer) in skippable_layers:
                layer = layer.inbound_nodes[0].inbound_layers
        non_skippable_inbounds.append(layer)
    return [ak_model.get_layer(ly.name) for ly in non_skippable_inbounds]



# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to set/get Akida target version"""

__all__ = ["AkidaVersion", "get_akida_version", "set_akida_version"]

import os
from enum import Enum
from contextlib import contextmanager


class AkidaVersion(Enum):
    v1 = "v1"
    v2 = "v2"


TARGET_AKIDA_VERSION = "CNN2SNN_TARGET_AKIDA_VERSION"


def get_akida_version():
    """Get the target akida version for model conversion.

    Returns:
        AkidaVersion: the target akida version, by default ``AkidaVersion.v2``
    """
    ak_str_version = os.environ.get(TARGET_AKIDA_VERSION, "v2")
    ak_version = getattr(AkidaVersion, ak_str_version, None)
    if ak_version is None:
        raise ValueError(
            f"{TARGET_AKIDA_VERSION}={ak_str_version} must be one of {AkidaVersion._member_names_}")
    return ak_version


@contextmanager
def set_akida_version(version):
    """Select the target akida version for model conversion.

    Args:
        version (AkidaVersion): the target Akida version.
    """
    assert isinstance(version, AkidaVersion), "Version must be an AkidaVersion"

    _prev_state = os.environ.get(TARGET_AKIDA_VERSION, None)
    try:
        os.environ[TARGET_AKIDA_VERSION] = version.value
        yield
    finally:
        # Recover default value
        if _prev_state is not None:
            os.environ[TARGET_AKIDA_VERSION] = _prev_state
        else:
            os.environ.pop(TARGET_AKIDA_VERSION)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedConv2D to Akida InputConvolutional or InputConv2D layer.
"""

from quantizeml.layers import QuantizedConv2D, QuantizedReLU, QuantizedMaxPool2D
from akida import InputConvolutional, InputConv2D

from ..akida_versions import AkidaVersion
from .activations import set_relu_variables
from .pooling import max_pool_param_checks
from .outputs import set_output_v1_variables, set_output_v2_variables
from .padding import check_conv_and_max_pool_compatibility
from .blocks import get_block_out_quantizer
from .block_converter import BlockConverter, register_conversion_patterns
from .conv_common import (parse_conv_block_v1, parse_conv_block_v2, get_layer_by_type,
                          set_conv_variables_v1, set_conv_variables_v2)


__all__ = ["InputConvBlockConverterV1", "InputConvBlockConverterV2"]

_PATTERNS = [(QuantizedConv2D, QuantizedReLU), (QuantizedConv2D, QuantizedMaxPool2D, QuantizedReLU)]


def convert_inputconv_block_v1(model_ak, block):
    """Converts a convolutional block into an akida InputConvolutional layer.

    The expected sequence is:

    - QuantizedConv2D,
    - QuantizedMaxPooling2D (optional),
    - QuantizedReLU.

    Args:
        model_ak (:obj:`akida.Model`): the target Akida model.
        block (list(:obj:`tf.keras.Layer`)): the block layers.

    Returns:
        bool: Returns True for a successful conversion.
    """

    conv = block[0]

    # Evaluate the convolutional block layers parameters
    layer_params = parse_conv_block_v1(block, input_layer=True)

    conv_ak = InputConvolutional(**layer_params)

    # Add layer to the model to build its internal variables
    model_ak.add(conv_ak)
    # Set base variables
    set_conv_variables_v1(conv_ak, conv, input_layer=True)

    # Get out_quantizer of the block.
    out_quantizer = get_block_out_quantizer(block)

    if out_quantizer:
        set_output_v1_variables(conv_ak, out_quantizer)

    return True


def convert_inputconv_block_v2(model_ak, block):
    """Converts a convolutional block into an akida InputConv2D layer.

    The expected sequence is:

    - QuantizedConv2D,
    - QuantizedMaxPooling2D (optional),
    - QuantizedReLU.

    Args:
        model_ak (:obj:`akida.Model`): the target Akida model.
        block (list(:obj:`tf.keras.Layer`)): the block layers.

    Returns:
        bool: Returns True for a successful conversion.
    """

    conv = block[0]

    # Evaluate the convolutional block layers parameters
    layer_params = parse_conv_block_v2(block, input_layer=True)

    conv_ak = InputConv2D(**layer_params)

    # Add layer to the model to build its internal variables
    model_ak.add(conv_ak)
    # Set base variables
    set_conv_variables_v2(conv_ak, conv, input_layer=True)

    # Set the activation variables
    relu_layer = get_layer_by_type(block, QuantizedReLU)
    set_relu_variables(conv_ak, relu_layer)

    # Get out_quantizer of the block.
    out_quantizer = get_block_out_quantizer(block)

    if out_quantizer:
        set_output_v2_variables(conv_ak, out_quantizer)

    return True


class InputConvBlockConverterV1(BlockConverter):
    """Main class that should be used to check if the input conv block is compatible to an Akida v1
    conversion and provides a method to convert it in an equivalent Akida InputConvolutional layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        self._inputconv_additional_checks()

    def _inputconv_additional_checks(self):
        conv = self._block[0]
        # Get conv layer input shape without the batch size.
        input_shape = conv.input_shape[1:]
        # InputConvolutional handles only image-like inputs (1-D or 3-D uint8 inputs)
        if input_shape[-1] not in (1, 3):
            raise ValueError("InputConvolutional layer can only handle image like inputs (i.e 3-D "
                             f"or 1-D input shape). Received: {input_shape}.")

        # The only weight bitwidth supported is [4, 8]
        weight_bits = conv.weight_quantizer.bitwidth
        if weight_bits not in [4, 8]:
            raise ValueError("InputConvolutional layer can only handle weights with 4 or 8 bits."
                             f" Received: {weight_bits}.")

        next_layer = self._block[1]
        if isinstance(next_layer, QuantizedMaxPool2D):
            check_conv_and_max_pool_compatibility(conv, next_layer)

    def convert(self, model_ak):
        assert len(model_ak.layers) == 0, "InputConvolutional should be the first model layer."
        return convert_inputconv_block_v1(model_ak, self._block)


class InputConvBlockConverterV2(BlockConverter):
    """Main class that should be used to check if the input conv block is compatible to an Akida v2
    conversion and provides a method to convert it in an equivalent Akida InputConv2D layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        self._inputconv_additional_checks()

    def _inputconv_additional_checks(self):
        conv = self._block[0]
        next_layer = self._block[1]
        if isinstance(next_layer, QuantizedMaxPool2D):
            check_conv_and_max_pool_compatibility(conv, next_layer)
            max_pool_param_checks(next_layer)

        # The only weight bitwidth supported is [4, 8]
        weight_bits = conv.weight_quantizer.bitwidth
        if weight_bits not in [4, 8]:
            raise ValueError("InputConvolutional layer can only handle weights with 4 or 8 bits."
                             f" Received: {weight_bits}.")

        # Make sure the Conv2D layers spatial params are square
        assert conv.kernel_size[0] == conv.kernel_size[1], ("InputConv2D handle only"
                                                            "square kernels")
        assert conv.strides[0] == conv.strides[1], ("InputConv2D stride should be"
                                                    "the same for both dimensions")

    def convert(self, model_ak):
        assert len(model_ak.layers) == 0, "InputConv2D should be the first model layer."
        return convert_inputconv_block_v2(model_ak, self._block)


# Register the valid input conv block pattern for Akida v1 and v2
register_conversion_patterns(AkidaVersion.v1, _PATTERNS, InputConvBlockConverterV1, True)
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, InputConvBlockConverterV2, True)



# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Helper functions to convert to akida models, models quantized with quantizeml API.
"""
from .model_generator import generate_model
from .block_converters_generator import generate_block_converters
from .dense import *
from .convolution import *
from .separable_convolution import *
from .conv2d_transpose import *
from .depthwise_conv2d import *
from .depthwise_conv2d_transpose import *
from .batchnorm import *
from .add import *
from .attention import *
from .shiftmax import *
from .madnorm import *
from .concatenate import *
from .extract_token import *
from .dequantizer import *
from .buffer_temp_conv import *
from .depthwise_buffer_temp_conv import *
from .stateful_recurrent import *



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert Stem block quantized layers to Akida.
Those layers are:
    - The Embedding layer
    - The Reshape layer
    - The ClassToken (+ the DistToken for distilled models) layer(s)
    - The AddPosEmbedding layer
"""
from akida import LayerType, Stem
from quantizeml.layers import (QuantizedConv2D, QuantizedReshape,
                               QuantizedClassToken, QuantizedAddPositionEmbs, OutputQuantizer)
import numpy as np
from .blocks import get_block_out_quantizer
from .outputs import set_output_v2_variables
from .weights import broadcast_and_set_variables
from .block_converter import BlockConverter, register_conversion_patterns
from ..akida_versions import AkidaVersion

_PATTERNS = [(QuantizedConv2D, QuantizedReshape, QuantizedClassToken, QuantizedAddPositionEmbs),
             (QuantizedConv2D, QuantizedReshape, QuantizedAddPositionEmbs),
             (QuantizedConv2D, QuantizedReshape, QuantizedClassToken, QuantizedClassToken,
              QuantizedAddPositionEmbs)]


def _get_cls_layers(block):
    """Helper that returns the QuantizedClassToken layers of a stem block.

    Args:
        block (list): the quantizeml quantized layers of the Stem to convert.

    Returns:
        list: Returns a list of the QuantizedClassToken layers of a stem block.
    """

    return [layer for layer in block if isinstance(layer, QuantizedClassToken)]


def _set_stem_variables(ak_layer, block):
    """Computes and sets the variables for an Akida Stem layer.

    This function converts the variables of a Keras layers and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        ak_layer (:obj:`akida.Layer`): the targeted akida layer.
        block (list): list of the source quantized layers block.
    """
    assert ak_layer.parameters.layer_type == LayerType.Stem

    embedding_layer = block[0]

    # Get the optional QuantizedClassToken layer(s) and the QuantizedAddPositionEmbs layer
    cls_layers = _get_cls_layers(block)
    add_pos_emb_layer = block[-1]

    # Prepare a dict for akida variables
    variables_ak = {}

    # get the Embedding weights
    weights_ak = embedding_layer.weight_quantizer.qweights.value.fp.values.numpy()

    # get the Embedding bias
    bias_quantizer = embedding_layer.bias_quantizer
    bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
    bias_shift = bias_quantizer.shift.value.numpy().astype(np.uint8)
    bias_ak = (bias >> bias_shift).astype(np.int8)

    if len(cls_layers) > 0:
        tokens_ak = []
        shifts_tokens_ak = []
        for cls_layer in cls_layers:
            # get the ClassToken layer token variable (aka cls member)
            cls_quantizer = cls_layer.cls_quantizer
            token = cls_quantizer.qweights.value.values.numpy().astype(np.int32)
            token_shift = cls_quantizer.shift.value.numpy().astype(np.uint8)
            token_ak = (token >> token_shift).astype(np.int8)
            # Insert the token value at the first position. This allows us to match
            # the model concatenation order.
            tokens_ak.insert(0, np.squeeze(token_ak))
            shifts_tokens_ak.insert(0, token_shift)
        variables_ak["tokens"] = np.stack(tokens_ak)
        variables_ak["tokens_shift"] = np.concatenate(shifts_tokens_ak)

    # get the positional embedding matrix
    pos_emb_quantizer = add_pos_emb_layer.pe_quantizer
    pos_emb = pos_emb_quantizer.qweights.value.values.numpy().astype(np.int32)
    pos_emb_shift = pos_emb_quantizer.shift.value.numpy().astype(np.uint8)
    pos_emb_ak = pos_emb >> pos_emb_shift
    variables_ak["pos_embedding"] = pos_emb_ak.astype(np.int8)
    # Get the QuantizedAddPositionEmbs layer shifts
    variables_ak["pos_embs_shift"] = pos_emb_shift

    variables_ak["weights"] = weights_ak.astype(np.int8)
    variables_ak["bias"] = bias_ak
    # Get the Embedding layer shifts
    variables_ak["bias_shift"] = bias_shift.astype(np.uint8)

    broadcast_and_set_variables(ak_layer, variables_ak)


def _parse_stem_block_layers(block):
    """Parses the quantizeml quantized additional layers of the Stem block and returns the
    params to create the corresponding Akida Stem layer.

    Args:
        layers (list): the quantizeml quantized layers of the Stem to convert.

    Returns:
        dict: the corresponding akida parameters.
    """
    # Extract neural layer of the block
    embedding = block[0]

    # Get the QuantizedClassToken layer(s)
    cls_layers = _get_cls_layers(block)

    # A block of layers always end with an OutputQuantizer. Extract it.
    out_quantizer = get_block_out_quantizer(block)
    assert isinstance(out_quantizer, OutputQuantizer)

    return dict(input_shape=embedding.input_shape[1:],
                filters=embedding.filters,
                kernel_size=embedding.kernel_size[0],
                buffer_bits=embedding.buffer_bitwidth + 1,
                output_bits=out_quantizer.bitwidth,
                num_non_patch_tokens=len(cls_layers),
                name=embedding.name)


def convert_quantized_stem_layers(model_ak, block):
    """Converts stem block layers and their variables and adds them to the
    Akida's model.

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the model will be added.
        block (list): the quantizeml quantized layers of the Stem to convert.

    Returns:
        bool: Returns True for a successful conversion.
    """
    # Parse the Stem block layers parameters
    stem_params = _parse_stem_block_layers(block)

    layer_ak = Stem(**stem_params)
    model_ak.add(layer_ak)
    _set_stem_variables(layer_ak, block)
    # Get out_quantizer of the block.
    out_quantizer = get_block_out_quantizer(block)
    set_output_v2_variables(layer_ak, out_quantizer)

    return True


class StemBlockConverter(BlockConverter):
    """Main class that should be used to check if the stem block is compatible to an Akida v2
    conversion and provides a method to convert it in an equivalent Akida stem layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        self._stem_additional_checks()

    def convert(self, model_ak):
        assert len(model_ak.layers) == 0, "Stem layer should be the first model layer."
        return convert_quantized_stem_layers(model_ak, self._block)

    def _stem_additional_checks(self):
        embedding = self._block[0]
        # Get embedding layer input shape without the batch size.
        input_shape = embedding.input_shape[1:]
        # Stem handles only square inputs
        if (input_shape[0] != input_shape[1]):
            raise ValueError("input should have square spatial dimensions. Receives: "
                             f"x_size={input_shape[0]} and y_size={input_shape[1]}")
        # Stem handles only image-like inputs (1-D or 3-D uint8 inputs)
        if input_shape[-1] not in (1, 3):
            raise ValueError("Stem layer can only handle image like inputs (i.e 3-D or 1-D input "
                             f"shape). Received: {embedding.input_shape[1:]}.")
        # Get the kernel size
        kernel_size = embedding.kernel_size
        # Stem handles only square kernels.
        if (kernel_size[0] != kernel_size[1]):
            raise ValueError("input should have square kernels. Receives: "
                             f"k_x={kernel_size[0]} and k_y={kernel_size[1]}")
        # Stem input size must be a multiple of kernel size.
        if (input_shape[0] % kernel_size[0] != 0):
            raise ValueError("Input spatial dimension size must be a multiple of kernel size")


# Register the valid stem block pattern for Akida v2
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, StemBlockConverter, input_pattern=True)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Helper functions to convert (Depthwise)BufferTempConv layers to their equivalent Akida layers.
"""
import numpy as np
from quantizeml.layers import (QuantizedReLU, WeightQuantizer,
                               AlignedWeightQuantizer, QuantizedBufferTempConv)
from .conv_common import get_layer_by_type
from .activations import parse_relu_v2, set_relu_variables
from .weights import broadcast_and_set_variable
from .blocks import get_block_out_quantizer
from .outputs import set_output_v2_variables, parse_output_bits, parse_post_op_buffer_bits


def parse_buf_temp_conv_block(block, depthwise=False):
    """Parses a quantizeml (depthwise) buffer_temp_conv block parameters for Akida v2.

    Args:
        block (list(:obj:`tf.keras.Layer`)): the buffer_temp_conv block layers.
        depthwise (bool, optional): boolean to declare the main layer as a depthwise layer.
            Defaults to False.

    Returns:
        dict: the corresponding Akida parameters.
    """

    btc_layer = block[0]

    # In quantizeml one bit is reserved for the sign in the buffer bitwidth
    # variable, but in akida this value has to be added back to have the
    # correct clipping.
    buffer_bits = btc_layer.buffer_bitwidth + 1

    block_params = dict(
        kernel_size=btc_layer.kernel_size,
        buffer_bits=buffer_bits,
        activation=False,
        name=btc_layer.name
    )
    # Add filters parameters if not a Depthwise BTC
    if not depthwise:
        assert isinstance(btc_layer, QuantizedBufferTempConv)
        block_params["filters"] = btc_layer.filters

    relu_layer = get_layer_by_type(block, QuantizedReLU)
    if relu_layer:
        act_params = parse_relu_v2(relu_layer)
        block_params.update(act_params)

    # parse the block output bits
    parse_output_bits(block, block_params)
    # parse the block post op buffer bits
    parse_post_op_buffer_bits(block, block_params)

    return block_params


def set_buffer_temp_conv_block_variables(layer_ak, block, depthwise=False):
    """Computes and sets the variables for an Akida v2 BufferTempConv layers.

    This function converts the variables of a Keras layers block and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        block (list(:obj:`tf.keras.Layer`)): the buffer temporal convolution block layers.
        depthwise (bool, optional): boolean to declare the main layer as a depthwise layer.
            Defaults to False.
    """
    btc_layer = block[0]

    assert isinstance(btc_layer.weight_quantizer, WeightQuantizer)
    if btc_layer.use_bias:
        assert isinstance(btc_layer.bias_quantizer, AlignedWeightQuantizer)

    # Get the weights
    weights = btc_layer.weight_quantizer.qweights.value.fp.values.numpy()
    if depthwise:
        assert len(weights.shape) == 2, "expected depthwise layer weights with a shape (T, C)"
        # Expand weights shape to match Akida expectation
        weights = weights.reshape((1, 1, *weights.shape))
    layer_ak.variables["weights"] = weights.astype(np.int8)

    # Set the bias (if there is one)
    if btc_layer.use_bias:
        bias_quantizer = btc_layer.bias_quantizer
        bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
        bias_shift = bias_quantizer.shift.value.numpy().astype(np.uint8)

        # Unshift the bias and store it
        layer_ak.variables["bias"] = (bias >> bias_shift).astype(np.int8)
        # Also store the bias shift
        broadcast_and_set_variable(layer_ak.variables, "bias_shift", bias_shift)

    # Set input shift if available
    if getattr(btc_layer, 'input_shift', None):
        broadcast_and_set_variable(layer_ak.variables, "input_shift",
                                   btc_layer.input_shift.value.numpy().astype(np.uint8))

    # Check if we have ReLU
    relu_layer = get_layer_by_type(block, QuantizedReLU)
    # Set optional activation variables
    if relu_layer:
        set_relu_variables(layer_ak, relu_layer)

    # Get the layer block output_quantizer
    out_quantizer = get_block_out_quantizer(block)
    # Set optional output_quantizer variables
    if out_quantizer:
        set_output_v2_variables(layer_ak, out_quantizer)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to set weights from quantizeml quantized Keras models to Akida.
"""
import numpy as np


def broadcast_and_set_variable(variables, var_name, value):
    """Adapts variables to akida variables shapes if necessary, and sets them.

    Args:
        variable (:obj:`akida.Variables`): the targeted akida variables.
        var_name (str): name of the variable
        value (int): value of the variable
    """
    var = variables[var_name]
    # If the variable is a scalar, broadcast it across the target akida
    # variable's last dimension.
    shape = var.shape
    if np.size(value) == 1:
        variables[var_name] = np.full(shape, value)
    else:
        variables[var_name] = np.reshape(value, shape)


def broadcast_and_set_variables(layer_ak, variables):
    """Adapts variables to akida variables shapes if necessary, and sets them.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        variables (dict): dictionary of variables
    """
    for var_name in variables:
        value = variables[var_name]
        broadcast_and_set_variable(
            layer_ak.variables, var_name, value)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["set_weight_variables"]

import numpy as np
from onnx import numpy_helper as np_onnx

from . import onnx_graph_tools


def broadcast_and_set_variable(variables, var_name, value):
    """Adapts variables to akida variables shapes if necessary, and sets them.

    Args:
        variable (akida.Variables): the targeted akida variables.
        var_name (str): then name of the variable
        new_value (int or np.ndarray): the new value of the variable
    """
    if not isinstance(value, np.ndarray):
        value = np.array(value)
    old_var = variables[var_name]
    # If the variable is a scalar, broadcast it across the target akida variable's last dimension
    new_var = np.full(old_var.shape, value) if value.size == 1 else value
    # Then, reshape and cast with expect shape and dtype, respectively
    variables[var_name] = new_var.reshape(old_var.shape).astype(old_var.dtype)


def to_value_shift(x, bitwidth=8, signed=True):
    """Transform a numpy array into a 8bit shift approximation.

    Args:
        x (np.ndarray): the array to transform
        bitwidth (int, optional): the maximal value representation. Defaults to 8.
        signed (bool, optional): whether the representation is signed or not. Defaults to True.

    Returns:
        np.ndarray, np.ndarray: the value and shift to represent input as value << shift.
    """
    if signed:
        # Reserve one bit to the sign
        bitwidth -= 1
    # Calculate integer bits, adding an epsilon in order to fix expect closed power of two
    # (e.g. 4 needs 3 integers and not np.ceil(np.log2(4)) == 2)
    # Note: we replace 0 by 2**bitwidth to avoid divide by zero
    y = np.where(x == 0, 2**bitwidth, x)
    int_bits = np.ceil(np.log2(np.abs(y)) + 1e-6)
    # Shift always have to be positive in this representation
    shift = np.maximum(0, int_bits - bitwidth)
    value = np.round(x / 2.0 ** shift)
    # Clip value to representable range
    value = np.clip(value, -2 ** bitwidth if signed else 0, 2 ** bitwidth - 1)
    return value, shift


def set_weight_variables(ak_layer, kernels, bias=None):
    """Set weights into akida variables.

    Args:
        ak_layer (akida.Layer): the akida layer to set variables.
        kernels (list of np.ndarray): the weights.
        bias (np.ndarray): if applies, the bias. Defaults to None.
    """
    # Set weights without broadcast: reshape/transpose must happen at conversion time
    ak_layer.variables["weights"] = kernels[0]
    if bias is not None:
        # Bias must be converted to a fixed point in the format: signed 8bit + unsigned shift
        ak_bias, ak_bias_shift = to_value_shift(bias)
        broadcast_and_set_variable(ak_layer.variables, "bias", ak_bias)
        broadcast_and_set_variable(ak_layer.variables, "bias_shift", ak_bias_shift)


def load_weights(node, initializers, func=None):
    """Parse model initializers that match with node inputs into numpy arrays.

    Args:

        node (NodeProto): the input node.
        initializers (list of TensorProto): the initializer list.
        func (FunctionProto, optional): the function that represents the node operation.
            If provided, the initializer names will have the inputs of the function
            instead of theirs own ones. Default None.

    Returns:
        dict: the initializers that were found.
    """
    fnames = func.input if func else node.input
    assert len(fnames) == len(node.input), "Mismatch between function and node inputs"

    # Align weight names with function input names
    weights = {}
    for name, fname in zip(node.input, fnames):
        weight = onnx_graph_tools.find_by_name(name, initializers)
        if weight:
            weights[fname] = np_onnx.to_array(weight)
    return weights



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["check_weight_types"]

import numpy as np


def check_weight_types(converter, weight_names, expect_types):
    """Check that weights have the expected type.

    Args:
        converter (OnnxConverter): the converter.
        weight_names (list of str): list of weight names to check.
        expect_types (list of str): list of expected weight types.

    Raises:
        RuntimeError: if a weight does not exist.
        ValueError: if a weight does not have the expected type.
    """
    assert len(weight_names) == len(expect_types), "Expect to have same lenght of elements."

    weights = converter.weights
    for w_name, e_type in zip(weight_names, expect_types):
        if w_name not in weight_names:
            raise RuntimeError(f"{converter.name} was expected to have '{w_name}' weight.")
        w_dtype = weights[w_name].dtype
        if w_dtype != np.dtype(e_type):
            raise ValueError(f"{w_name} in {converter.name} must be {e_type} type, not {w_dtype}.")


def check_attributes(converter, attr_names):
    """Check that attributes were loaded into converter.

    Args:
        converter (OnnxConverter): the converter.
        weight_names (list of str): list of attribute names to check.

    Raises:
        RuntimeError: if an attribute does not exist.
    """
    for expect_attr in attr_names:
        if not hasattr(converter, expect_attr):
            raise RuntimeError(f"{converter.name} expect to have '{expect_attr}' attribute.")


def check_if_squared(value, name_val=None):
    """Check if input value is square.

    Args:
        value (object): the value to check.
        name_val (str, optional): name of the value. Defaults to None.

    Raises:
        ValueError: if value is not square.
    """
    assert hasattr(value, "__iter__")
    if name_val is None:
        name_val = str(value)
    if value[0] != value[1]:
        raise ValueError(f"{name_val} is expected to be square.")



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedAttention to Akida.
"""
from akida import Attention
from quantizeml.layers import QuantizedAttention
import numpy as np

from ..akida_versions import AkidaVersion
from .weights import broadcast_and_set_variable
from .layer_utils import get_inbound_layers
from .block_converter import BlockConverter, register_conversion_patterns
from .blocks import get_block_out_quantizer
from .outputs import set_output_v2_variables, parse_output_bits, parse_post_op_buffer_bits


__all__ = ["AttentionBlockConverter"]

_PATTERNS = [(QuantizedAttention,)]


def _set_attention_variables(ak_layer, block):
    """Computes and sets the variables for an Akida v2 Attention layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        ak_layer (:obj:`akida.Layer`): the targeted akida layer.
        block (list(:obj:`tf.keras.Layer`)): the block of keras layers.
    """
    attention_layer = block[0]
    variables_ak = ak_layer.variables

    # Get the QuantizedAttention layer shifts
    value_shift = attention_layer.values_shift.value.numpy().astype(np.uint8)
    broadcast_and_set_variable(variables_ak, "value_shift", value_shift)
    shiftmax = attention_layer.softmax_op
    shiftmax_input_shift = shiftmax.input_shift.value.numpy()
    broadcast_and_set_variable(ak_layer.variables, "shiftmax_input_shift",
                               shiftmax_input_shift.astype(np.uint8))
    out_quantizer = get_block_out_quantizer(block)
    if out_quantizer:
        set_output_v2_variables(ak_layer, out_quantizer)


def _create_attention(block):
    """Parses a quantizeml QuantizedAttention layer and returns the params to
    create the corresponding Akida Attention layer.

    Args:
        block (list(:obj:`tf.keras.Layer`)): list of quantizeml quantized layers.

    Returns:
        :obj:`akida.Attention`: The created akida layer.
    """
    attention_layer = block[0]
    # In quantizeml one bit is reserved automatically for the sign, but in akida
    # this is rather checked during the clipping operations.
    block_params = {"buffer_bits": attention_layer.buffer_bitwidth + 1}

    # parse the block output bits
    parse_output_bits(block, block_params)

    # parse the block post op buffer bits
    parse_post_op_buffer_bits(block, block_params)

    # As with output_bits and buffer_bits, shiftmax_output_bits is set with one
    # bit more.
    block_params['shiftmax_output_bits'] = attention_layer.softmax_op.exp_bitwidth + 1

    block_params['num_heads'] = attention_layer.num_heads
    return Attention(**block_params,
                     name=attention_layer.name)


def convert_quantized_attention(model_ak, block):
    """Converts QuantizedAttention layer and its variables and adds it to the
    Akida's model.

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the model will be added.
        block (list(:obj:`tf.keras.Layer`)): list of quantizeml quantized layers.

    Returns:
        bool: Returns True for a successful conversion.
    """
    layer_ak = _create_attention(block)
    # Retrieve the akida inbound layers
    inbound_layers_ak = get_inbound_layers(model_ak, block[0])
    model_ak.add(layer_ak, inbound_layers_ak)
    _set_attention_variables(layer_ak, block)
    return True


class AttentionBlockConverter(BlockConverter):
    """Main class that should be used to check if the attention layer block is compatible to an
    Akida v2 conversion and provides a method to convert it in an equivalent Akida Attention layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        self._attention_additional_checks()

    def _attention_additional_checks(self):
        shiftmax = self._block[0].softmax_op
        # if input_shift available, make sure it has no negative values
        if shiftmax.input_shift.value:
            shiftmax_input_shift = shiftmax.input_shift.value.numpy()
            if np.any(shiftmax_input_shift < 0):
                raise ValueError(
                    f"Layer {self._block[0].name} contains negative values for " +
                    "shiftmax_input_shift, that is not supported")

    def convert(self, model_ak):
        return convert_quantized_attention(model_ak, self._block)


# Register the valid attention block pattern for Akida v2
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, AttentionBlockConverter)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedAdd to Akida.
"""
from akida import Add
import quantizeml.layers as qlayers
import numpy as np

from ..akida_versions import AkidaVersion
from .layer_utils import get_inbound_layers
from .weights import broadcast_and_set_variable
from .block_converter import BlockConverter, register_conversion_patterns
from .blocks import get_block_out_quantizer
from .outputs import set_output_v2_variables, parse_output_bits, parse_post_op_buffer_bits


__all__ = ["AddBlockConverter"]

_PATTERNS = [(qlayers.QuantizedAdd,)]


def _set_add_variables(ak_layer, block):
    """Computes and sets the variables for an Akida Add layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        ak_layer (:obj:`akida.Layer`): the targeted akida layer.
        block (list(:obj:`tf.keras.Layer`)): the block of keras layers.
    """
    add_layer = block[0]

    variables_ak = ak_layer.variables
    a_shift = add_layer.a_shift.value.numpy().astype(np.uint8)
    broadcast_and_set_variable(variables_ak, "a_shift", a_shift)
    b_shift = add_layer.b_shift.value.numpy().astype(np.uint8)
    broadcast_and_set_variable(variables_ak, "b_shift", b_shift)

    out_quantizer = get_block_out_quantizer(block)
    if out_quantizer:
        set_output_v2_variables(ak_layer, out_quantizer)


def _create_add(block):
    """Parses a quantizeml QuantizedAdd layer and returns the corresponding
    Akida v2 Add layer.

    Args:
        block (list(:obj:`tf.keras.Layer`)): list of quantizeml quantized layers.

    Returns:
        :obj:`akida.Add`: The created akida layer.
    """
    add_layer = block[0]
    # In quantizeml one reserves automaticaly one bit for the sign, but in akida
    # this is rather checked during the clipping operations.
    add_params = {"buffer_bits": add_layer.buffer_bitwidth + 1}
    # parse the block output bits
    parse_output_bits(block, add_params)
    # parse the block post op buffer bits
    parse_post_op_buffer_bits(block, add_params)

    return Add(**add_params,
               name=add_layer.name)


def convert_quantized_add_block(model_ak, block):
    """Converts QuantizedAdd layer block and its variables and adds it to the
    Akida's v2 model.

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the model will be added.
        block (list(:obj:`tf.keras.Layer`)): list of quantizeml quantized layers.

    Returns:
        bool: Returns True for a successful conversion.
    """
    # Retrieve the akida inbound layers
    inbound_layers_ak = get_inbound_layers(model_ak, block[0])
    # Create and add layer to the akida model
    layer_ak = _create_add(block)
    model_ak.add(layer_ak, inbound_layers_ak)
    # Set the akida layer converted variables
    _set_add_variables(layer_ak, block)
    return True


class AddBlockConverter(BlockConverter):
    """Main class that should be used to check if the Add layer block is compatible to an
    Akida v2 conversion and provides a method to convert it in an equivalent Akida
    Add layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def convert(self, model_ak):
        return convert_quantized_add_block(model_ak, self._block)


# Register the valid add block pattern for Akida v2
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, AddBlockConverter)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["AddOnnxConverter"]

import numpy as np

import akida

from .base_converter import OnnxConverter
from .register import register_onnx_converter_target
from .weights import broadcast_and_set_variable
from .check_compatibility import check_weight_types
from .scale_out import set_output_scale_variables


@register_onnx_converter_target("QuantizedAdd")
class AddOnnxConverter(OnnxConverter):
    """Convert QuantizedAdd type node into akida.Add.

    Args:
        node (NodeProto): the node to convert.
        model (ModelProto): the model containing the node.
    """

    def load_attributes(self, node):
        # Some attributes should infer from node.op_type
        n_op = node.op_type
        self.out_scaled = "Scaled" in n_op
        return super().load_attributes(node)

    def _additional_checks(self):
        # Expect weights/attributes
        check_weight_types(self, ["Xs", "Ys"], ["int32", "int32"])
        if self.out_scaled:
            check_weight_types(self, ["Shift"], ["float32"])

        # Check input shifts are always positive
        if np.any(self.weights["Xs"] < 0) or np.any(self.weights["Ys"] < 0):
            raise RuntimeError(f"Impossible to convert {self.name} into akida. "
                               "Input shifts should be positive.")

        # Reject Add(X, X) (inputs are the same tensor)
        if self._node.input[0] == self._node.input[1]:
            raise RuntimeError(f"Impossible to convert {self.name} into akida. "
                               "Inputs must come from different tensors.")

    def _parse_akida_layer(self):
        return akida.Add(output_bits=8 if self.out_scaled else 32, name=self.name)

    def _set_akida_variables(self, ak_layer):
        def to_power(x):
            return np.log2(x)

        assert isinstance(ak_layer, akida.Add)

        # Input shifts
        ak_vars = ak_layer.variables
        broadcast_and_set_variable(ak_vars, "a_shift", to_power(self.weights["Xs"]))
        broadcast_and_set_variable(ak_vars, "b_shift", to_power(self.weights["Ys"]))

        # Shift out
        if self.out_scaled:
            set_output_scale_variables(ak_layer, shift=self.weights["Shift"])



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Utility function referent to padding.
"""
import numpy as np

import quantizeml

from ..model_generator import _get_padding

quantized_conv = (quantizeml.layers.QuantizedConv2D,
                  quantizeml.layers.QuantizedDepthwiseConv2D,
                  quantizeml.layers.QuantizedSeparableConv2D)


def get_padding(k_layer):
    """From a convolution layer, return the akida.Padding.

    Args:
        k_layer (:obj:`tf.keras.Layer`): the source quantized layer.

    Returns:
        tuple: ``akida.Padding``.
    """
    assert isinstance(k_layer, quantized_conv)

    padding = k_layer.padding
    padding_ak_value = get_padding_value(k_layer)
    # If a padding_value value is provided or layer produces same output with 'same' or 'valid',
    # force the padding to be "same" in Akida.
    if np.any(padding_ak_value != 0) or check_same_valid_compatibility(k_layer):
        padding = "same"
    return _get_padding(padding)


def get_padding_value(k_layer):
    """From a convolution layer, return the padding value.

    Args:
        k_layer (:obj:`tf.keras.Layer`): the source quantized layer.

    Returns:
        tuple: ``akida.Padding`` and the array of padding values.
    """
    assert isinstance(k_layer, quantized_conv)

    # Padding value must be built in constructor
    padding_ak_value = np.array(0, "int32")
    if getattr(k_layer, "padding_value", None):
        padding_quantize = k_layer.padding_value_quantizer
        assert isinstance(padding_quantize, quantizeml.layers.AlignedWeightQuantizer)
        padding_ak_value = padding_quantize.qweights.value.values.numpy().astype(np.int32)
    if np.any(padding_ak_value < 0) or np.any(padding_ak_value > 255):
        raise ValueError("padding_value must be an uint8 value")
    return padding_ak_value


def check_same_valid_compatibility(layer):
    """Check if a layer produces the same output regardless of its padding ('same' or 'valid').

    Args:
        layer (:obj:`akida.Layer`): Layer to verify

    Returns:
        bool: same/valid compatibility result.
    """
    # Layer produces same output when kernel size == 1
    return layer.kernel_size == (1, 1)


def check_conv_and_max_pool_compatibility(conv_layer, max_pool_layer):
    """Check if a conv layer and the following max_pool_layer are compatible for a conversion.
    Raise an error if not.

    Args:
        conv_layer (:obj:`akida.Layer`): QuantizedConv2D layer to verify.
        max_pool_layer (:obj:`akida.Layer`): QuantizedMaxPool2D layer to verify.

    """

    neur_pad = conv_layer.padding if not check_same_valid_compatibility(conv_layer) else "same"
    pool_pad = getattr(max_pool_layer, "padding", "")
    if neur_pad != pool_pad:
        raise ValueError(f"Pooling layer {max_pool_layer.name} (padding: {pool_pad}) must"
                         f" have the same padding as {conv_layer.name} (padding: {neur_pad}).")



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Parsing functions to generate an Akida model from a Keras model quantized with quantizeml api.
"""

from quantizeml.models import record_quantization_variables
from akida import Model
from onnx import ModelProto

from .input_data import convert_input
from .block_converters_generator import generate_block_converters
from .block_converter import BlockConverter
from .stem import StemBlockConverter
from .input_conv import InputConvBlockConverterV1, InputConvBlockConverterV2
from .onnx_conversion import generate_onnx_model


def _raise_block_error(block):
    """ Raise error due to non convertible layers block"""
    block_layers_name = ""
    block_layers_type = ""
    for layer in block:
        block_layers_name += f"{layer.name}, "
        block_layers_type += f"{layer.__class__.__name__}, "
    raise RuntimeError(f"Layers {block_layers_name[:-2]}: unsupported type "
                       f"[{block_layers_type[:-2]}].")


def generate_model(model):
    """Generates an Akida model.

    This function creates an Akida model by iterating through the layers of the
    quantized model. For each layer, the corresponding akida layer is created and
    added to the Akida model.

    Args:
        model (:obj:`tf.keras.Model` or :obj:`onnx.ModelProto`): a model to convert.

    Returns:
        :obj:`akida.Model`: the generated Akida model.
    """
    if isinstance(model, ModelProto):
        return generate_onnx_model(model)
    return generate_qml_model(model)


def generate_qml_model(model):
    """Generates an Akida model based on a Keras quantizeml model.

    Args:
        model (:obj:`tf.keras.Model`): a model to convert.

    Returns:
        :obj:`akida.Model`: the generated Akida model.
    """
    # Check if input model is convertible and extract its layers blocks
    blocks = generate_block_converters(model)
    # First store necessary variables for conversion
    record_quantization_variables(model)
    model_ak = Model()

    # Flag to identify if the current block has been converted
    converted = False
    # Input blocks are handled with their associated BlockConverter
    if isinstance(blocks[0], (StemBlockConverter, InputConvBlockConverterV1,
                              InputConvBlockConverterV2)):
        converted = blocks[0].convert(model_ak)
    if not converted:
        # Convert the keras InputLayer into an InputData layer
        convert_input(model_ak, model)
    else:
        blocks.pop(0)

    for block in blocks:
        # The next check converts the potential layers block to its corresponding Akida layer
        if isinstance(block, BlockConverter):
            if block.convert(model_ak):
                continue
        # If you got here, the block is not supported: raise an error.
        _raise_block_error(block)
    return model_ak



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Custom tf.keras.constraint.Constraint object to limit weights minimum value.
"""

import tensorflow as tf

from keras.constraints import Constraint


class MinValueConstraint(Constraint):
    """ Constraint that ensures that weights values are not below a minimum
    value.

    Args:
        min_value: the minimum desired value for the weights
    """

    def __init__(self, min_value=1e-2):
        self.min_value = min_value

    def __call__(self, w):
        return tf.clip_by_value(w, self.min_value, tf.float32.max)

    def get_config(self):
        config = super().get_config()
        config.update({'min_value': self.min_value})
        return config



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import numpy as np

import akida

from .base_converter import OnnxConverter
from .check_compatibility import check_weight_types, check_attributes, check_if_squared
from .weights import broadcast_and_set_variable, set_weight_variables
from .activation import set_activation_variables
from .scale_out import set_output_scale_variables


def set_convolutional_variables(converter, ak_layer, is_depthwise=False):
    """Transfer converter weights to ak_layer.

    Args:
        converter (OnnxConverter): the converter to extract weights.
        ak_layer (akida.Layer): the target Akida model.
        is_depthwise (bool, optional): whether to transpose kernel on depthwise. Defaults to False.
    """
    assert isinstance(converter, OnnxConverter)

    # Set padding value (in case of InputConv2D)
    if ak_layer.parameters.layer_type == akida.LayerType.InputConv2D:
        ak_variables = ak_layer.variables
        broadcast_and_set_variable(ak_variables, "padding_value", converter.weights["x_pad_value"])

    # Get kernels and transpose them (FCKxKy -> KxKyCF)
    kernel = converter.weights["W"].transpose((2, 3, 1, 0))
    if not converter.is_input_layer:
        # Kernel need to be flipped
        kernel = np.flip(kernel, axis=(0, 1))

    # Akida interprets depthwise kernel with respect to ONNX in a different way:
    # * Akida: Number of filters is 1, with C input channels.
    # * Onnx: Channels is 1 (C / groups) and filters is the desired output F.
    # That is why we need to transpose both dimensions
    if is_depthwise:
        kernel = kernel.transpose((0, 1, 3, 2))

    bias = converter.weights.get("bias", None)
    set_weight_variables(ak_layer, [kernel], bias)

    # Activation
    if converter.activation:
        set_activation_variables(ak_layer, converter.weights.get("max_value", None))

    # Scale out
    set_output_scale_variables(ak_layer, converter.weights["Scale"], converter.weights["Shift"])


def check_convolution_compatibility(converter, use_pad_value=False):
    """Check convolution compatibility with Akida.

    Args:
        converter (OnnxConverter): the converter to check.
        use_pad_value (bool, optional): extend checks to padding value. Defaults to False.
    """
    # Weights constrains
    check_weight_types(converter,
                       weight_names=["W", "Scale", "Shift", "pads"],
                       expect_types=["int8", "uint8", "float32", "int64"])

    # Zero point constrains
    if use_pad_value:
        check_weight_types(converter, ["x_pad_value"], expect_types=["uint8"])
        channel = converter.weights["W"].shape[1]
        pad_size = converter.weights["x_pad_value"].size
        if pad_size != channel:
            raise ValueError("Padding value size must be equal to the number of channel.")
        if not converter.is_input_layer and np.any(converter.weights["x_pad_value"]):
            raise ValueError("Padding value must be zero.")

    # Bias constrains
    if converter.use_bias:
        check_weight_types(converter=converter, weight_names=["bias"], expect_types=["int32"])

    # Max value constrains
    if "max_value" in converter.weights:
        check_weight_types(converter=converter, weight_names=["max_value"], expect_types=["int32"])

    # Attributes checks
    check_attributes(converter, ["strides"])
    if converter.pool_type == akida.PoolType.Max:
        check_attributes(converter, ["pool_size", "pool_strides", "pool_pads"])

    # Other checks
    kernel_shapes = converter.weights["W"].shape[-2:]
    check_if_squared(kernel_shapes, name_val="kernels")
    check_if_squared(converter.strides, name_val="strides")
    if converter.pool_type == akida.PoolType.Max:
        check_if_squared(converter.pool_size, name_val="pool sizes")
        check_if_squared(converter.pool_strides, name_val="pool strides")



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedDepthwiseConv2DTranspose to Akida.
"""

from quantizeml.layers import QuantizedDepthwiseConv2DTranspose, QuantizedReLU
from akida import DepthwiseConv2DTranspose

from ..akida_versions import AkidaVersion
from .block_converter import BlockConverter, register_conversion_patterns
from .conv_common import parse_conv2d_transpose_block, set_conv2d_transpose_block_variables

__all__ = ["DepthwiseConvTransposeBlockConverter"]

_PATTERNS = [(QuantizedDepthwiseConv2DTranspose,),
             (QuantizedDepthwiseConv2DTranspose, QuantizedReLU)]


def convert_depthwise_conv2d_transpose_block(model_ak, block):
    """Converts a depthwise convolutional transpose block into an akida v2 DepthwiseConv2DTranspose
    layer.

    The expected sequence is:

    - QuantizedDepthwiseConv2DTranspose,
    - QuantizedReLU (optional).

    Args:
        model_ak (:obj:`akida.Model`): the target Akida model.
        block (list(:obj:`tf.keras.Layer`)): the block layers.

    Returns:
        bool: Returns True for a successful conversion.
    """

    # Evaluate the depthwise convolutional transpose layer parameters
    conv_params = parse_conv2d_transpose_block(block, depthwise=True)

    # Create Akida layer
    dw_conv_ak = DepthwiseConv2DTranspose(**conv_params)
    # Add layer to the model to build its internal variables
    model_ak.add(dw_conv_ak)

    # Set base variables
    set_conv2d_transpose_block_variables(dw_conv_ak, block)

    return True


class DepthwiseConvTransposeBlockConverter(BlockConverter):
    """Main class that should be used to check if the depthwise conv transpose block is compatible
    to an Akida v2 conversion and provides a method to convert it in an equivalent Akida
    DepthwiseConv2DTranspose layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        self._depthwise_conv_additional_checks()

    def _depthwise_conv_additional_checks(self):
        depth_transpose = self._block[0]

        # Make sure the DepthwiseConv2DTranspose kernel size and stride params are square
        assert depth_transpose.kernel_size[0] == depth_transpose.kernel_size[1], (
            "DepthwiseConv2DTranspose kernel should be square")
        assert depth_transpose.strides[0] == depth_transpose.strides[1], (
            "DepthwiseConv2DTranspose strides should be the same on both dimension")
        assert depth_transpose.kernel_size[0] in [3, 4], (
            "DepthwiseConv2DTranspose supports kernel size of 3 or 4 only")
        assert depth_transpose.strides[0] == 2, "DepthwiseConv2DTranspose supports stride 2 only"

        # Padding value must be same
        assert depth_transpose.padding == "same", ("DepthwiseConv2DTranspose handles only 'same'"
                                                   " padding")

        # The only weight bitwidth supported is [4, 8]
        weight_bits = depth_transpose.weight_quantizer.bitwidth
        assert weight_bits in [4, 8], ("DepthwiseConv2DTranspose layer can only handle weights"
                                       f" with 4 or 8 bits. Received: {weight_bits}.")

    def convert(self, model_ak):
        return convert_depthwise_conv2d_transpose_block(model_ak, self._block)


# Register the valid depthwise conv transpose block pattern for Akida v2
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, DepthwiseConvTransposeBlockConverter)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["InputDataOnnxConverter"]

import akida

from .base_converter import OnnxConverter
from .register import register_onnx_converter_target


def get_next_neighbor_nodes(node, graph):
    # Get the nodes connected to the outputs
    outbounds = []
    for target_node in graph.node:
        for target_input in target_node.input:
            if target_input in node.output:
                outbounds.append(target_node)
    return outbounds


@register_onnx_converter_target("InputQuantizer")
class InputDataOnnxConverter(OnnxConverter):
    """Convert InputData node into an akida.InputData.

    Args:
        node (NodeProto): the node to convert.
        model (ModelProto): the model that the node is.
    """
    @property
    def is_input_layer(self):
        return True

    @property
    def input_name(self):
        return self._node.output[0]

    @property
    def func(self):
        return None

    def load_attributes(self, node):
        self.input_signed = True
        self.input_bits = 8

    def _additional_checks(self):
        assert len(self._node.output) == 1, "InputData must have exactly one output."
        # Check input data is connected to only one node
        outbound_nodes = get_next_neighbor_nodes(self._node, self._model.graph)
        if len(outbound_nodes) != 1:
            raise RuntimeError(f"{self.name} must be connected to only one node.")

    def _parse_akida_layer(self):
        return akida.InputData(input_shape=self.input_shape,
                               input_bits=self.input_bits,
                               input_signed=self.input_signed,
                               name=self.name)

    def _set_akida_variables(self, ak_layer):
        # Nothing to do
        ...



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["OnnxConverter"]

import numpy as np

from quantizeml.onnx_support.layers.base_layer import DOMAIN

from . import onnx_graph_tools
from .layer_bounds import get_inbound_layers
from .weights import load_weights


def get_akida_input_model_shape(onnx_graph):
    tensor_name = onnx_graph.input[0].name
    shape = onnx_graph_tools.get_input_info(onnx_graph)[tensor_name][0]
    # Remove batch size as akida requires a 3D input shape
    akida_shape = shape[1:]
    if len(akida_shape) > 1:
        # Convert to channels last
        akida_shape = akida_shape[1:] + akida_shape[0:1]
    # Expand to have exactly three dimensions
    akida_shape = np.insert(akida_shape, [0] * (3 - len(akida_shape)), 1)
    return akida_shape


class OnnxConverter:
    """Abstract class that allows to convert a node into the corresponding Akida layer.

    Child should overwrite _additional_checks function (if extra checks are required)
    and implement convert() function.

    Args:
        node (NodeProto): the node to convert.
        model (ModelProto): the model that has the node in it.
    """

    def __init__(self, node, model):
        assert node.domain == DOMAIN, f"Unrecognized {node.name}: it is not part of the domain."
        self._node = node
        self._model = model
        self._input_shape = None

        # Parse attributes and weights from node (and initializer) into class
        self.weights = load_weights(node, model.graph.initializer, self.func)
        self.load_attributes(node)

        # Check special rules
        self._additional_checks()

    @property
    def name(self):
        return self._node.name

    @property
    def input_name(self):
        return self._node.input[0]

    @property
    def input_shape(self):
        if self.is_input_layer:
            return get_akida_input_model_shape(self._model.graph)
        assert self._input_shape is not None
        return self._input_shape

    @property
    def is_input_layer(self):
        # Default to False
        return False

    @property
    def func(self):
        func = onnx_graph_tools.find_by_name(self._node.op_type, self._model.functions)
        assert func, f"{self.name} does not have an associated function."
        return func

    def load_attributes(self, node):
        """Load node attributes into object.

        Args:
            node (NodeProto): the input node.
        """
        for attr in node.attribute:
            # Get attribute value
            value = onnx_graph_tools.get_field(node, attr.name)
            # Set value into class as an attribute
            setattr(self, attr.name, value)

    def _additional_checks(self):
        """Check node compatibility with Akida."""
        ...

    def _parse_akida_layer(self):
        raise NotImplementedError("Child must implement this function")

    def _set_akida_variables(self, ak_layer):
        raise NotImplementedError("Child must implement this function")

    def convert(self, ak_model):
        """Convert node into an Akida layer and append it into the model.

        Args:
            ak_model (akida.Model): the target Akida model.
        """
        if len(ak_model.layers) > 0 and self.is_input_layer:
            raise RuntimeError(f"Impossible to convert {self.name} into akida. "
                               "There can be no layers prior to it.")
        if len(ak_model.layers) == 0 and not self.is_input_layer:
            raise RuntimeError(f"Impossible to convert {self.name} into akida. "
                               "This node require an InputData.")

        # Retrieve the akida inbound layers
        inbound_layers_ak = get_inbound_layers(ak_model, self._node, self._model.graph)

        # Assign input shape
        input_shape = [x.output_dims for x in inbound_layers_ak]
        if len(input_shape) == 1:
            input_shape = input_shape[0]
        self._input_shape = input_shape

        # Declare akida layer to set variables
        ak_layer = self._parse_akida_layer()

        # Append new layer in akida mode
        ak_model.add(ak_layer, inbound_layers_ak)

        # Align akida layer name with convert one
        self._node.name = ak_layer.name

        # Set weights into akida layer
        self._set_akida_variables(ak_layer)



import quantizeml.onnx_support.graph_tools as onnx_graph_tools
import quantizeml.onnx_support.layers.subgraph_ops as onnx_subgraph_ops

from .input_data import *
from .conv2d import *
from .depthwise2d import *
from .dense import *
from .add import *
from .dequantizer import *
from .model_generator import generate_onnx_model



from os import path
from setuptools import setup
from platform import platform


def get_tf_dep():
  platform_string = platform()
  if 'macOS' in platform_string and 'arm64' in platform_string:
    tf_name = 'tensorflow-macos'
  else:
    tf_name = 'tensorflow'
  return tf_name + '>=2.10.0,<2.13.0'


# Read the contents of the README file
directory = path.abspath(path.dirname(__file__))
with open(path.join(directory, 'README'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='cnn2snn',
    version='2.7.0',
    description='Keras to Akida CNN Converter',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Nezar Lheimeur',
    author_email='nlheimeur@brainchip.com',
    url='https://doc.brainchipinc.com',
    license='Apache 2.0',
    packages=['cnn2snn', 'cnn2snn.transforms', 'cnn2snn.calibration', 'cnn2snn.quantizeml',
              'cnn2snn.quantizeml.onnx_conversion'],
    entry_points={
        'console_scripts': [ 'cnn2snn = cnn2snn.cli:main' ]
    },
    install_requires=[get_tf_dep(), 'keras>=2.10.0,<2.13.0',
        'akida==2.7.0', 'quantizeml~=0.8.1'],
)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2019 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert weights from Keras to Akida.

"""
from collections import namedtuple
import numpy as np
import tensorflow as tf
import keras.layers as layers
from akida import LayerType

KerasLayerParams = namedtuple(
    'KerasLayerParams',
    ['bias', 'act_threshold', 'act_step', 'global_avg_pool_factor'])


def set_layer_variables(layer_ak,
                        input_scaling,
                        layer_neural,
                        layer_activation=None,
                        layer_globalavgpool=None):
    """Computes and sets weights/thresholds for an Akida layer.

    This function converts the weights and variables of a sequence of Keras
    layers (a neural layer, an optional global average pooling layer and an
    optional activation layer) into the weights and thresholds for the given
    Akida layer.

    The input scaling corresponds either to the input scaling of the model if
    the Akida layer is the first layer, or the input scaling of the previous
    activation layer if the Akida layer is an intermediate or final layer.
    """

    # Convert weights
    weights, weights_scale_factor = \
        convert_layer_weights(layer_ak.parameters.layer_type, layer_neural,
                              layer_ak.input_dims)

    # Convert fire thresholds and steps
    th_fire, th_step = _convert_layer_thresholds(layer_ak.parameters.layer_type,
                                                 layer_neural, layer_activation,
                                                 layer_globalavgpool,
                                                 input_scaling,
                                                 weights_scale_factor, weights)

    # Set weights
    if layer_ak.parameters.layer_type == LayerType.SeparableConvolutional:
        layer_ak.set_variable('weights', weights[0])
        layer_ak.set_variable('weights_pw', weights[1])
    else:
        layer_ak.set_variable('weights', weights)

    # Set thresholds
    layer_ak.set_variable('threshold', th_fire)
    layer_ak.set_variable("act_step", th_step)


def convert_layer_weights(layer_type, layer_k, input_dims=None):
    """Converts weights for an Akida layer.

    This function returns the converted weights for any Akida layer,
    according to the Akida layer type and the Keras neural layer.

    Args:
        layer_type (str): type of the layer Akida.
        layer_k (:obj:`tf.keras.Layer`): the Keras layer from which the weights
            will be converted.
        input_dims (tuple): input dimensions of the Keras layer. This argument
            is only required for a fullyConnected Akida layer.

    Returns:
        tuple: the converted weights and the weights scale factor from
            quantization.

    """
    if layer_type == LayerType.InputConvolutional:
        return _convert_conv_weights(layer_k, input_conv=True)
    if layer_type == LayerType.Convolutional:
        return _convert_conv_weights(layer_k)
    if layer_type == LayerType.SeparableConvolutional:
        return _convert_separable_conv_weights(layer_k)
    if layer_type == LayerType.FullyConnected:
        return _convert_dense_weights(layer_k, input_dims)
    assert False, "The layer type is unmanaged."
    return tuple()


def _convert_conv_weights(layer, input_conv=False):
    # Quantize weights
    weights = layer.get_weights()[0]
    wq, delta = quantize_weights(layer.quantizer, weights)

    wq_akida = wq * delta
    wq_akida = np.round(wq_akida).astype(np.int8)

    # Flip W and H dimensions for conv. kernels (not input conv.)
    if not input_conv:
        wq_akida = np.flip(wq_akida, axis=[0, 1])

    return wq_akida, delta


def _convert_separable_conv_weights(layer):
    # Quantize depthwise weights
    weights = layer.get_weights()[0]
    wq_dw, delta_dw = quantize_weights(layer.quantizer_dw, weights)
    w_ak_dw = wq_dw * delta_dw
    w_ak_dw = np.round(w_ak_dw).astype(np.int8)
    w_ak_dw = np.flip(w_ak_dw, axis=[0, 1])

    # Quantize pointwise weights
    weights_pw = layer.get_weights()[1]
    wq_pw, delta_pw = quantize_weights(layer.quantizer, weights_pw)
    # Pointwise weights in Keras have HWCN format and H = W = 1. This
    # makes the conversion to Akida's NCHW trivial.
    w_ak_pw = wq_pw * delta_pw
    w_ak_pw = np.round(w_ak_pw).astype(np.int8)

    # For separable, the weights scale factor can be seen as the product of the
    # DW and PW scale factors.
    delta = delta_dw * delta_pw
    return (w_ak_dw, w_ak_pw), delta


def _convert_dense_weights(layer, input_dims):
    # Quantize weights
    weights = layer.get_weights()[0]
    wq, delta = quantize_weights(layer.quantizer, weights)

    # retrieve input dimensions from Akida's layer
    inX, inY, inchans = input_dims
    # When a fully connected layer follows a convolutional layer, we need to modify
    # the way the weights are laid out because the Keras and Akida flatten operations
    # that happen on spatial dimensions are inverted (row-major versus col-major).
    # We therefore need to:
    # - reshape the Keras (FxN) weights to match the input shape (XxYxCxN),
    # - transpose to the equivalent akida shape (CxYxXxN),
    # - reshape back to a flat shape (1x1xFxN).
    #
    wq_akida = wq.reshape(inX, inY, inchans, layer.units) \
        .transpose(2, 1, 0, 3) \
        .reshape(1, 1, inchans * inY * inX, layer.units)
    # Multiply by delta, round and cast to int
    wq_akida = wq_akida * delta
    wq_akida = np.round(wq_akida).astype(np.int8)

    return wq_akida, delta


def _prepare_threshold_calculations(layer_ak_type,
                                    layer_neural,
                                    layer_activation=None,
                                    layer_globalavgpool=None):
    """Retrieves required variables to calculate Akida fire threshold and
    step.

    This function returns a dictionary with variables required for threshold
    calculations of the given Akida layer map: bias in the Keras neural layer,
    parameters of the activation layer and global average pooling factor.

    Args:
        layer_ak_type (:obj:`akida.LayerType`): the type of the Akida layer.
        layer_neural (:obj:`tf.keras.Layer`): the Keras neural layer.
        layer_activation (:obj:`tf.keras.Layer`): the Keras activation layer.
        layer_globalavgpool (:obj:`tf.keras.Layer`): the optional Keras global
            average pooling layer.

    Returns:
        dict: required variables to calculate Akida fire threshold and step.
    """

    # Initialize variables to their default values
    threshold_shape = layer_neural.output_shape[-1]
    bias = np.zeros(threshold_shape)
    thresh_keras = np.zeros(threshold_shape)
    step_keras = np.ones(threshold_shape)
    global_avg_pool_factor = 1

    # Get bias if present
    if layer_neural.use_bias:
        if layer_ak_type == LayerType.SeparableConvolutional:
            bias = layer_neural.get_weights()[2]
        else:
            bias = layer_neural.get_weights()[1]

    # Get parameters from the discrete ReLU activation (threshold and step)
    if layer_activation:
        thresh_keras = (layer_activation.threshold.numpy() * np.ones(threshold_shape))
        step_keras = (
            (layer_activation.step.numpy() *
             2.**layer_activation.bitwidth /
             16) *
            np.ones(threshold_shape))

    # Get global average pooling factor
    if isinstance(layer_globalavgpool, layers.GlobalAveragePooling2D):
        global_avg_pool_factor = np.prod(layer_globalavgpool.input_shape[1:3])

    return KerasLayerParams(bias, thresh_keras, step_keras,
                            global_avg_pool_factor)


def _convert_layer_thresholds(layer_ak_type, layer_neural, layer_activation,
                              layer_globalavgpool, input_scaling,
                              weights_scale_factor, weights_ak):
    """Returns the fire thresholds and threshold steps for a given Akida layer.

    This function computes the fire thresholds and steps for a given Akida layer
    from the corresponding neural, activation and global avg pooling Keras
    layers. Other arguments are required to compute these variables, such as
    the input scaling, the weights scale factor, and the Akida weights (only
    used for an InputConvolutional Akida layer).

    Args:
        layer_ak_type (:obj:`akida.LayerType`): the type of the Akida layer.
        layer_neural (:obj:`tf.keras.Layer`): the Keras neural layer.
        layer_activation (:obj:`tf.keras.Layer`): the Keras activation layer.
        layer_globalavgpool (:obj:`tf.keras.Layer`): the optional Keras global
            average pooling layer.
        input_scaling (2-element tuple): the input scaling of the current layer.
        weights_scale_factor (float): the weights scale factor of the current
            Akida layer. This is given by the 'convert_layer_weights' function.
        weights_ak (np.ndarray): the Akida weights of the current Akida layer.

    Returns:
        tuple: the fire thresholds and steps.

    """

    input_scale, input_shift = input_scaling
    p = _prepare_threshold_calculations(layer_ak_type, layer_neural,
                                        layer_activation, layer_globalavgpool)

    # Compute threshold fires and steps
    th_fire = (input_scale * weights_scale_factor * (p.act_threshold - p.bias))
    th_step = (input_scale * weights_scale_factor * p.act_step).astype(
        np.float32)

    # If there is an InputConv layer, update th_fire with input shift
    if layer_ak_type == LayerType.InputConvolutional:
        th_fire += np.sum(weights_ak, axis=(0, 1, 2)) * input_shift

    th_fire *= p.global_avg_pool_factor
    th_step *= p.global_avg_pool_factor

    th_fire = np.floor(th_fire).astype(np.int32)

    return th_fire, th_step


def quantize_weights(quantizer, w):
    """Returns quantized weights and delta as numpy arrays.

    Internally, it uses a tf.function that wraps calls to the quantizer in
    a graph, allowing the weights to be quantized eagerly.

    Args:
        quantizer (:obj:`WeightQuantizer`): the quantizer object.
        w (:obj:`np.ndarray`): the weights to quantize.

    Retruns:
        :obj:`np.ndarray`: the quantized weights `np.ndarray` and the scale
            factor scalar.

    """

    w_tf = tf.constant(w)
    wq = quantizer.quantize(w_tf)
    scale_factor = quantizer.scale_factor(w_tf)
    return wq.numpy(), scale_factor.numpy()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["generate_onnx_model"]
import warnings

from onnx.helper import make_node
from onnx.checker import check_model
from copy import deepcopy

import akida

from .base_converter import DOMAIN
from .register import map_node_to_converter


def get_convertible_nodes(model):
    """Return the node list of candidates to convert.

    Args:
        model (ModelProto): the model to convert.

    Returns:
        list of NodeProto: the node list.
    """
    # Check if model needs to add/remove an InputQuantizer
    min_layers = 2
    if len(model.graph.node) > 0:
        if model.graph.node[0].op_type == "InputQuantizer" and len(model.graph.node) > 1:
            first_layer = map_node_to_converter(model.graph.node[1], model)
            if first_layer.is_input_layer:
                # Remove InputQuantizer when next layer is a HRC
                model.graph.node.pop(0)
        else:
            first_layer = map_node_to_converter(model.graph.node[0], model)
            if not first_layer.is_input_layer:
                # Insert a fake InputQuantizer to convert it into InputData
                input_names = [x.name for x in model.graph.input]
                fake_iq = make_node("InputQuantizer", [], input_names, domain=DOMAIN)
                model.graph.node.insert(0, fake_iq)
        if model.graph.node[0].op_type != "InputQuantizer":
            # Update minimum layers required : [HRC]
            min_layers = 1

    # Check if there is a Dequantizer at the end of the graph.
    # Nodes after that will be skipped from conversion.
    idx = 0
    for idx, node in enumerate(model.graph.node):
        if node.op_type == "Dequantizer":
            break

    # Check model is not empty (at least one node to convert is required)
    if len(model.graph.node[: idx + 1]) < min_layers:
        raise ValueError("Model is empty or does not have any node to convert.")

    # Prints a warning message with a summary of the skipped nodes
    skip_nodes = model.graph.node[idx + 1:]
    if len(skip_nodes) > 0:
        stop_layer_msg = " at node " + model.graph.node[idx - 1].name if idx > 0 else ""
        skip_layers_summary = "___________________________________________________\n"
        skip_layers_summary += "Node (type)\n"
        skip_layers_summary += "===================================================\n"
        for node in skip_nodes:
            skip_layers_summary += node.name + " (" + node.op_type + ")\n"
        skip_layers_summary += "===================================================\n"
        warnings.warn("Conversion stops" + stop_layer_msg + " because of a dequantizer. "
                      "The end of the graph is ignored:\n" + skip_layers_summary)

    return model.graph.node[: idx + 1]


def generate_onnx_model(model):
    """Generates an Akida model based on an ONNX quantizeml model.

    Args:
        model (obj:`onnx.ModelProto`): a ONNX model to convert.

    Returns:
        akida.Model: the generated Akida model.
    """
    # Clone model to keep the original intact
    model = deepcopy(model)

    # Model must be compatible with ONNX
    check_model(model, full_check=True)

    # Checks over model
    assert len(model.graph.input) == 1, "Unsupported mode: it must have exactly one input."
    assert len(model.graph.output) == 1, "Unsupported mode: it must have exactly one output."

    # Now create akida model and iterate nodes to convert each one in an akida layer.
    target_nodes = get_convertible_nodes(model)
    akida_model = akida.Model()
    for node in target_nodes:
        converter = map_node_to_converter(node, model)
        converter.convert(akida_model)
    return akida_model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2019 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Helper to update V1 models act_step if too high to be mappable"""

import warnings
import numpy as np
import akida

# act_step should fit in 16bit
ACT_STEP_HW_LIMIT = (1 << 16) - 1
THRESHOLD_HW_LIMIT = (1 << 19) - 1


def fix_v1_activation_variables(model):
    """ Helper to equalize the Akida model act_step if those are not mappable.

    If the act_step of a layer is higher the value allowed in hardware, this method will update
    this value and modify the model to adapt it to the limit. To do that, it will compute the ratio
    by which the act_step should be divide to fit in the HW limit should be divide to fit in the
    HW limit and then divide the weights of the next layer by the next ratio to keep the
    equivalence.

    Args:
        model (`akida.Model`): an Akida model.

    """
    # This operation is compatible only for akida v1 target
    if model.ip_version != akida.core.IpVersion.v1:
        return
    # Equalize model act_steps
    _equalize_act_step(model)

    # Saturate thresholds
    _saturate_threshold(model)


def _equalize_act_step(model):
    for id, layer in enumerate(model.layers[:-1]):
        if "act_step" in layer.variables.names:
            act_step = layer.variables["act_step"]
            if (act_step > ACT_STEP_HW_LIMIT).any():
                original_act_step = act_step.copy()
                # Compute the ratio by which the act_step should be divide to fit in the
                # HW requirement
                ratio = np.where(act_step > ACT_STEP_HW_LIMIT, act_step / ACT_STEP_HW_LIMIT, 1)
                # Update the non valid act_steps
                act_step[act_step > ACT_STEP_HW_LIMIT] = ACT_STEP_HW_LIMIT
                layer.variables["act_step"] = act_step
                # If the layer has an activation, the threshold should also be updated
                # since it holds a part of the act_step. To do so we assume that the Akida
                # model has been converted through the cnn2snn API
                if layer.parameters.activation:
                    act_bits = layer.parameters.act_bits
                    # Remove from the threshold the effect of the original act_step,
                    # then update with the new one
                    threshold_effect = np.round(0.5 * (act_step - original_act_step))
                    threshold_effect /= 2 ** (act_bits - 4)
                    layer.variables["threshold"] += threshold_effect.astype(np.int32)
                # get the next layer weights
                next_weights = model.get_layer(id + 1).variables["weights"]
                # divide the next layer weights by the ratio to keep the equivalence
                next_weights = next_weights // ratio[np.newaxis, np.newaxis, :, np.newaxis]
                model.get_layer(id + 1).variables["weights"] = next_weights.astype(np.int8)


def _saturate_threshold(model):
    for layer in model.layers[:-1]:
        if "threshold" in layer.variables.names:
            threshold = layer.variables["threshold"]
            # Saturate the threshold value to the max threshold value supported by the HW
            # Note: that this might slightly affect the global model accuracy in comparison
            # with its quantized version.
            if np.any(threshold > THRESHOLD_HW_LIMIT):
                threshold[threshold > THRESHOLD_HW_LIMIT] = THRESHOLD_HW_LIMIT
                warnings.warn(f"Warning: The {layer.name} layer holds very high threshold \
                          values which are not compatible with the hardware. Those are clipped \
                          to the maximum supported value {THRESHOLD_HW_LIMIT}.",
                              category=UserWarning)
                layer.variables["threshold"] = threshold



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["Dense1DOnnxConverter"]

import numpy as np

import akida

from .base_converter import OnnxConverter
from .register import register_onnx_converter_target
from .weights import set_weight_variables
from .activation import set_activation_variables
from .scale_out import set_output_scale_variables
from .check_compatibility import check_weight_types


@register_onnx_converter_target("QuantizedDense1D")
class Dense1DOnnxConverter(OnnxConverter):
    """Convert QuantizedDense1D type node into akida.Dense1D.

    Args:
        node (NodeProto): the node to convert.
        model (ModelProto): the model that the node is.
    """

    def load_attributes(self, node):
        # Some attributes should infer from node.op_type
        n_op = node.op_type
        self.is_flatten = "Flatten" in n_op
        self.activation = "ReLU" in n_op
        self.use_bias = "Biased" in n_op
        self.out_scaled = "Scaled" in n_op
        return super().load_attributes(node)

    def _additional_checks(self):
        # Expect weights/attributes
        check_weight_types(converter=self, weight_names=["W"], expect_types=["int8"])
        if self.use_bias:
            check_weight_types(self, ["bias"], ["int32"])
        if "max_value" in self.weights:
            check_weight_types(self, ["max_value"], ["int32"])
        if self.out_scaled:
            check_weight_types(self, ["Scale", "Shift"], ["uint8", "float32"])

    def _parse_akida_layer(self):
        # Parse common information
        layer_params = {
            "name": self.name,
            "units": self.weights["W"].shape[0],
            "output_bits": 8 if self.out_scaled else 32,
            "activation": self.activation,
        }
        return akida.Dense1D(**layer_params)

    def _set_akida_variables(self, ak_layer):
        assert isinstance(ak_layer, akida.Dense1D)

        # Get kernels
        kernels = self.weights["W"]
        if self.is_flatten:
            # Akida and ONNX flatten operations are not identical.
            # We need to reshape weights to compensate.
            # Note: original input_shape is (c, x, y), but property return akida shape
            x, y, c = self.input_shape
            # First, unroll flattened inputs
            kernels = np.reshape(kernels, (-1, c, x, y))
            # Second, transpose to match akida ordering
            kernels = np.transpose(kernels, (2, 3, 1, 0))
            # Finally, flatten again
            kernels = np.reshape(kernels, (x * y * c, -1))
        else:
            # Transpose weights to match akida ordering (channel at last dimension)
            kernels = np.transpose(kernels)

        # Get bias
        bias = self.weights.get("bias", None)
        set_weight_variables(ak_layer, [kernels], bias)

        # Activation
        if self.activation:
            set_activation_variables(ak_layer, self.weights.get("max_value", None))

        # Scale out
        if self.out_scaled:
            set_output_scale_variables(ak_layer, self.weights["Scale"], self.weights["Shift"])



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Transformation utilities for Keras/CNN2SNN Sequential models.
"""

from keras import Model, Sequential, Input
from keras.layers import ReLU, Dropout, Concatenate, InputLayer, MaxPool2D

from ..quantization_layers import QuantizedActivation
from .clone import (clone_model_with_weights, clone_layer,
                    clone_layer_and_add_to_model)
from .batch_normalization import invert_batchnorm_pooling, fold_batchnorm


def prepare_to_convert(model):
    # Make sure the model is sequential
    seq_model = sequentialize(model)

    # For now, we support only models with a single branch
    if not isinstance(seq_model, Sequential):
        raise RuntimeError(
            "The model contains more than one sequential branch.")

    # Transform model to prepare conversion: change the order of layers,
    # fold BN, freeze quantizers, remove useless layers.
    sync_model = syncretize(seq_model)

    return sync_model


def sequential_remove_useless_layers(model):
    """Removes useless layers in a Sequential model.

    Useless layers are:
    - Dropout

    Args:
        model (:obj:`tf.keras.Model`): a Sequential Keras model.
    Returns:
        :obj:`tf.keras.Model`: a Sequential Keras model.
    """

    _raise_error_if_model_not_sequential(model)

    useless_layers = (Dropout,)

    new_model = Sequential()
    new_model.add(Input(model.input_shape[1:]))
    for layer in model.layers:
        if isinstance(layer, useless_layers):
            continue
        clone_layer_and_add_to_model(layer, new_model)

    return new_model


def sequential_invert_pooling_activation(model):
    """Invert activation and MaxPool2D layer in a Sequential model to have
    MaxPool2D before the activation.

    Having activation->MaxPool2D or MaxPool2D->activation is equivalent only if
    the activation is increasing (ok for ReLU and QuantizedActivation).
    Note that GlobalAvgPool2D cannot be inverted with an activation because
    there is no equivalence between activation->GAP and GAP->activation.

    Args:
        model (:obj:`tf.keras.Model`): a Sequential Keras model.
    Returns:
        :obj:`tf.keras.Model`: a Sequential Keras model.
    """

    _raise_error_if_model_not_sequential(model)

    new_model = Sequential()
    new_model.add(Input(model.input_shape[1:]))
    i = 0
    while i < len(model.layers) - 1:
        layer = model.layers[i]
        next_layer = model.layers[i + 1]
        if (isinstance(layer, (ReLU, QuantizedActivation)) and isinstance(next_layer, MaxPool2D)):
            clone_layer_and_add_to_model(next_layer, new_model)
            clone_layer_and_add_to_model(layer, new_model)
            i = i + 2
        else:
            clone_layer_and_add_to_model(layer, new_model)
            i = i + 1

    if i < len(model.layers):
        clone_layer_and_add_to_model(model.layers[-1], new_model)

    return new_model


def _raise_error_if_model_not_sequential(model):
    """Raises a ValueError if model is not a Sequential Keras model.
    """
    if not isinstance(model, Sequential):
        raise ValueError(f"Model is expected to be Sequential. Receives type "
                         f"{model.__class__.__name__}.")


def _check_layers_data_format(model):
    """Asserts that all layers in the model are 'channels_last'.
    Args:
        model (tf.keras.model): the Keras model to check.
    """

    # Error if 'data_format' is 'channels_first'
    for layer in model.layers:
        if hasattr(layer, 'data_format'):
            if layer.data_format == "channels_first":
                raise RuntimeError("Unsupported data format 'channels_first' "
                                   f"in layer '{layer.name}'.")


def _check_layers_dilation(model):
    """Asserts that all layers in the model have default dilation.
    Args:
        model (tf.keras.model): the Keras model to check.
    """

    # Error if 'dilation_rate' is not default
    for layer in model.layers:
        if hasattr(layer, 'dilation_rate') and layer.dilation_rate != (1, 1):
            raise ValueError(f"Unsupported dilation rate {layer.dilation_rate} in layer "
                             f"'{layer.name}'.")


def _check_model_input_output(model):
    """Asserts that model inputs/outputs are supported for conversion.
    The Keras model must have only one input layer and one output layer. The
    input shape must 4-D (N, H, W, C).

    Args:
        model (tf.keras.model): the Keras model to check.
    """

    # Error if multiple inputs
    if len(model.input_names) > 1:
        raise RuntimeError("Model must have only one input layer. Receives "
                           f"inputs {model.input_names}.")

    # Error if multiple outputs
    if len(model.output_names) != 1:
        raise RuntimeError("Model must have only one output layer. Receives"
                           f"outputs {model.output_names}.")

    # Error if input shape is not 4D, i.e. (N, H, W, C)
    if len(model.input_shape) not in (2, 4):
        raise RuntimeError(
            "Input shape of model must be 2-D or 4-D (batch size + 1-D or 3-D "
            f"tensors). Receives input shape {model.input_shape}.")


def _check_layer_inbounds(model):
    """Asserts that all layers in the model have only one inbound node and
    inbound layer (except for input layers and Concatenate layers).
    """
    for layer in model.layers:
        if len(layer.inbound_nodes) != 1:
            raise RuntimeError("Layers must have only one inbound node. "
                               f"Receives layer {layer.name} with "
                               f"{len(layer.inbound_nodes)} nodes.")

        if (layer.inbound_nodes[0].is_input or isinstance(layer, Concatenate)):
            continue

        num_inbound_layers = len(layer.inbound_nodes[0].parent_nodes)
        if num_inbound_layers != 1:
            raise RuntimeError(f"Layer {layer.name} must have only one inbound "
                               f"layer. Receives {num_inbound_layers}.")


def _check_concat_layer_compatibility(layer, inbound_layers):
    """Checks that the Concatenate layer is compatible for conversion:

    - A Concatenate layer axis parameter must be equal to -1.
    - A Concatenate layer must have exactly two inbound layers.

    Args:
        layer (:obj:`tf.keras.Layer`): a Keras Concatenate layer.
        inbound_layers (list): the inbound layers of the Concatenate layer.
    """

    if layer.axis != -1:
        raise RuntimeError(f"The Concatenate layer '{layer.name}' must have "
                           f"axis=-1. Receives axis={layer.axis}.")

    if len(inbound_layers) != 2:
        inbound_names = [layer.name for layer in inbound_layers]
        raise RuntimeError(f"The Concatenate layer '{layer.name}' must have "
                           f"exactly two inbound layers. Receives inbound "
                           f"layers {inbound_names}.")


def sequentialize(model):
    """Transform a Model into Sequential sub-models and Concatenate layers.

    This function returns an equivalent model where all linear branches are
    replaced by a Sequential sub-model.

    Args:
        model (:obj:`tf.keras.Model`): a Keras model

    Returns:
        :obj:`tf.keras.Model`: a Keras model with Sequential sub-models
    """

    # Clone model to avoid shared layers
    model_clone = clone_model_with_weights(model)

    _check_layers_data_format(model_clone)
    _check_layers_dilation(model)
    _check_model_input_output(model_clone)
    _check_layer_inbounds(model_clone)

    if isinstance(model_clone, Sequential):
        return model_clone

    def parse_layer(layer, visited_layers, current_branch, output_tensors):
        """Go through a TensorFlow/Keras graph by recursively looking at
        outbound layers.

        Each linear branch is detected and converted to a Sequential sub-model.
        A linear branch ends if:
          - the current layer has multiple outbounds (split connections)
          - the current layer has no outbound (output layer of the model)
          - the next layer is a Concatenate layer
        """

        if layer in visited_layers:
            raise RuntimeError(f"Layer {layer.name} already visited.")

        # Do not visit this layer if all inbound layers were not visited yet
        # (e.g. for Concatenate inbounds)
        inbound_layers = [n.layer for n in layer.inbound_nodes[0].parent_nodes]
        if set(inbound_layers).difference(visited_layers):
            return
        visited_layers.append(layer)

        # Skip input layer but store its output tensor for graph connection
        if isinstance(layer, InputLayer):
            output_tensors[layer] = layer.output
        # Add layer to graph if layer is Concatenate
        elif isinstance(layer, Concatenate):
            _check_concat_layer_compatibility(layer, inbound_layers)
            # Get input tensors of Concatenate layer
            concat_input_tensors = [output_tensors[layer] for layer in inbound_layers]
            # Add Concatenate layer to the graph
            output_tensors[layer] = layer(concat_input_tensors)

        else:
            # Add current layer to current branch
            current_branch.append(layer)

            # End current branch and add it to the graph if:
            # - current layer has multiple outbounds (split connections)
            # - current layer has no outbound (output layer of the model)
            # - next layer is Concatenate
            if (len(layer.outbound_nodes) != 1 or isinstance(
                    layer.outbound_nodes[0].layer, Concatenate)):

                parent_nodes = current_branch[0].inbound_nodes[0].parent_nodes
                input_tensor = output_tensors[parent_nodes[0].layer]

                # Create sub-model for current branch and add it to the graph
                submodel = Sequential(current_branch)
                output_tensors[layer] = submodel(input_tensor)
                current_branch.clear()

        # Go to next layer
        for next_layer in [node.layer for node in layer.outbound_nodes]:
            parse_layer(next_layer, visited_layers, current_branch,
                        output_tensors)

    # Go through model layers to detect Sequential branches
    input_layer = model_clone.get_layer(model_clone.input_names[0])
    output_layer = model_clone.get_layer(model_clone.output_names[0])
    output_tensors = {}
    parse_layer(input_layer,
                visited_layers=[],
                current_branch=[],
                output_tensors=output_tensors)

    # Create new model with Sequential sub-models
    model_cut = Model(model_clone.input, output_tensors[output_layer])

    # Clone model to avoid shared layers
    model_cut = clone_model_with_weights(model_cut)

    # Sanity check: the first layer should always be an InputLayer
    if not isinstance(model_cut.layers[0], InputLayer):
        raise ValueError("Incompatible model")
    # Subsequent layers are either Concatenate or Sequential
    for layer in model_cut.layers[1:]:
        if not isinstance(layer, (Concatenate, Sequential)):
            raise ValueError("Incompatible model")

    # If the Model contains a single branch, return it
    if len(model_cut.layers) == 2:
        return model_cut.layers[1]

    return model_cut


def syncretize(model):
    """Align all linear branches of a Model with akida layer sequences.

    The input model must be composed of Sequential submodels and Concatenate
    layers. This function will apply transformations on every Sequential
    submodel and returns an equivalent functional model with akida compatible
    sequences of layers.

    Args:
        model (:obj:`tf.keras.Model`): a Keras model with Sequential submodels.

    Returns:
        :obj:`tf.keras.Model`: a Keras model with akida-compatible Sequential
        submodels.
    """

    def syncretize_sequential(submodel):
        submodel = sequential_remove_useless_layers(submodel)
        submodel = sequential_invert_pooling_activation(submodel)
        submodel = invert_batchnorm_pooling(submodel)
        submodel = fold_batchnorm(submodel)
        return submodel

    if isinstance(model, Sequential):
        return syncretize_sequential(model)

    def parse_submodel(layer, visited_layers, output_tensors):
        if layer in visited_layers:
            raise RuntimeError(f"Layer {layer.name} already visited.")

        # Do not visit this layer if all inbound layers were not visited yet
        # (e.g. for Concatenate inbounds)
        inbound_layers = [n.layer for n in layer.inbound_nodes[0].parent_nodes]
        if set(inbound_layers).difference(visited_layers):
            return
        visited_layers.append(layer)

        # Add layer to graph if layer is InputLayer or Concatenate
        if isinstance(layer, InputLayer):
            input_clone = clone_layer(layer)
            output_tensors[layer] = input_clone.output
        elif isinstance(layer, Concatenate):
            # Get input tensors of Concatenate layer
            concat_input_tensors = [output_tensors[layer] for layer in inbound_layers]
            # Add Concatenate layer to the graph
            concat_clone = clone_layer(layer)
            output_tensors[layer] = concat_clone(concat_input_tensors)
        elif isinstance(layer, Sequential):
            # Get input tensors of submodel
            parent_nodes = layer.inbound_nodes[0].parent_nodes
            input_tensor = output_tensors[parent_nodes[0].layer]
            # Transform submodel and add it to the graph
            new_submodel = syncretize_sequential(layer)
            output_tensors[layer] = new_submodel(input_tensor)
        else:
            raise RuntimeError(f"Layer {layer.name} of type "
                               f"{layer.__class__.__name__} is not supported "
                               f"here.")

        # Go to next layer
        for next_layer in [node.layer for node in layer.outbound_nodes]:
            parse_submodel(next_layer, visited_layers, output_tensors)

    # Go through model layers to transform Sequential branches
    input_layer = model.get_layer(model.input_names[0])
    output_tensors = {}
    parse_submodel(input_layer,
                   visited_layers=[],
                   output_tensors=output_tensors)

    # Create new model with transformed Sequential sub-models
    output_layer = model.get_layer(model.output_names[0])
    return Model(output_tensors[input_layer], output_tensors[output_layer])



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedDepthwiseConv2D to Akida.
"""
import numpy as np

from quantizeml.layers import (QuantizedDepthwiseConv2D, QuantizedReLU,
                               QuantizedMaxPool2D, WeightQuantizer, AlignedWeightQuantizer)
from akida import DepthwiseConv2D

from ..akida_versions import AkidaVersion
from .activations import set_relu_variables
from .weights import broadcast_and_set_variable
from .outputs import set_output_v2_variables, parse_output_bits, parse_post_op_buffer_bits
from .padding import check_conv_and_max_pool_compatibility
from .pooling import max_pool_param_checks
from .blocks import get_block_out_quantizer
from .conv_common import get_layer_by_type, parse_depthwise_conv_block
from .block_converter import BlockConverter, register_conversion_patterns

__all__ = ["DepthwiseConvBlockConverter"]

_PATTERNS = [(QuantizedDepthwiseConv2D,), (QuantizedDepthwiseConv2D, QuantizedReLU),
             (QuantizedDepthwiseConv2D, QuantizedMaxPool2D, QuantizedReLU)]


def _set_depthwise_conv2d_block_variables(layer_ak, block):
    """Computes and sets the variables for an Akida v2 DepthwiseConv2D layer.

    This function converts the variables of a Keras layers block and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        block (list(:obj:`tf.keras.Layer`)): the depthwise convolutional block layers.
    """
    depth_conv = block[0]
    assert isinstance(depth_conv.weight_quantizer, WeightQuantizer)
    if depth_conv.use_bias:
        assert isinstance(depth_conv.bias_quantizer, AlignedWeightQuantizer)

    # Get the weights
    weights = depth_conv.weight_quantizer.qweights.value.fp.values.numpy()
    # Flip W and H dimensions for depthwise
    weights = np.flip(weights, axis=[0, 1])

    layer_ak.variables["weights"] = weights.astype(np.int8)

    # Set the bias (if there is one)
    if depth_conv.use_bias:
        bias_quantizer = depth_conv.bias_quantizer
        bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
        bias_shift = bias_quantizer.shift.value.numpy().astype(np.uint8)

        # Unshift the bias and store it
        layer_ak.variables["bias"] = (bias >> bias_shift).astype(np.int8)
        # Also store the bias shift
        broadcast_and_set_variable(layer_ak.variables, "bias_shift", bias_shift)

    # Set input shift
    input_shift = getattr(depth_conv, 'input_shift', None)
    if input_shift is not None:
        broadcast_and_set_variable(layer_ak.variables, "input_shift",
                                   depth_conv.input_shift.value.numpy().astype(np.uint8))

    # Check if we have ReLU
    relu_layer = get_layer_by_type(block, QuantizedReLU)
    if relu_layer:
        set_relu_variables(layer_ak, relu_layer)

    # Get the layer block output_quantizer
    out_quantizer = get_block_out_quantizer(block)
    # Set optional output_quantizer variables
    if out_quantizer:
        set_output_v2_variables(layer_ak, out_quantizer)


def convert_depthwise_conv_block(model_ak, block):
    """Converts a depthwise convolutional block into an akida v2 DepthwiseConv2D layer.

    The expected sequence is:

    - QuantizedDepthwiseConv2D,
    - QuantizedMaxPool2D (optional),
    - QuantizedReLU (optional).

    Args:
        model_ak (:obj:`akida.Model`): the target Akida model.
        block (list(:obj:`tf.keras.Layer`)): the block layers.

    Returns:
        bool: Returns True for a successful conversion.

    """

    # Evaluate the depthwise convolutional layer parameters
    conv_params = parse_depthwise_conv_block(block)

    # add output quantizer bitwidth parameter
    parse_output_bits(block, conv_params)
    # parse the block post op buffer bits
    parse_post_op_buffer_bits(block, conv_params)

    # Create Akida layer
    dw_conv_ak = DepthwiseConv2D(**conv_params)
    # Add layer to the model to build its internal variables
    model_ak.add(dw_conv_ak)

    # Set base variables
    _set_depthwise_conv2d_block_variables(dw_conv_ak, block)

    return True


class DepthwiseConvBlockConverter(BlockConverter):
    """Main class that should be used to check if the depthwise conv block is compatible to an
    Akida v2 conversion and provides a method to convert it in an equivalent Akida DepthwiseConv2D
    layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        self._depthwise_conv_additional_checks()

    def _depthwise_conv_additional_checks(self):
        depth_conv = self._block[0]
        # Make sure the DepthwiseConv2D kernel size and stride params are square
        assert depth_conv.kernel_size[0] == depth_conv.kernel_size[1], (
            "DepthwiseConv2D kernel should be square")
        assert depth_conv.strides[0] == depth_conv.strides[1], (
            "DepthwiseConv2D strides should be the same on both dimensions")

        # The only weight bitwidth supported is [4, 8]
        weight_bits = depth_conv.weight_quantizer.bitwidth
        assert weight_bits in [4, 8], ("DepthwiseConv2D layer can only handle weights"
                                       f" with 4 or 8 bits. Received: {weight_bits}.")

        # Check optional pooling compatibility
        pool_layer = get_layer_by_type(self._block, QuantizedMaxPool2D)
        if pool_layer:
            check_conv_and_max_pool_compatibility(depth_conv, pool_layer)
            max_pool_param_checks(pool_layer)

    def convert(self, model_ak):
        return convert_depthwise_conv_block(model_ak, self._block)


# Register the valid depthwise conv block pattern for Akida v2
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, DepthwiseConvBlockConverter)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert keras pooling layers parameters to akida.
"""

from quantizeml.layers import QuantizedMaxPool2D
from akida import Padding, PoolType


def parse_max_pool_v1(layer):
    """Parses a quantizeml.QuantizedMaxPool2D parameters for Akida v1.

    Args:
        layer (:obj:`quantizeml.QuantizedMaxPool2D`): the layer to parse.

    Returns:
        dict: the corresponding akida parameters.
    """
    assert isinstance(layer, QuantizedMaxPool2D)

    padding = Padding.Same if layer.padding == "same" else Padding.Valid
    pool_size = layer.pool_size
    pool_stride = layer.strides if layer.strides else pool_size

    pool_params = dict(
        pool_type=PoolType.Max,
        padding=padding,
        pool_size=pool_size,
        pool_stride=pool_stride
    )

    return pool_params


def parse_max_pool_v2(layer):
    """Parses a quantizeml.QuantizedMaxPool2D parameters for Akida v2.

    Args:
        layer (:obj:`quantizeml.QuantizedMaxPool2D`): the layer to parse.

    Returns:
        dict: the corresponding akida parameters.
    """
    assert isinstance(layer, QuantizedMaxPool2D)

    padding = Padding.Same if layer.padding == "same" else Padding.Valid
    pool_size = layer.pool_size
    pool_stride = layer.strides if layer.strides else pool_size

    pool_params = dict(
        pool_type=PoolType.Max,
        padding=padding,
        pool_size=pool_size[0],
        pool_stride=pool_stride[0]
    )

    return pool_params


def set_gap_variables_v1(layer_ak, gap_layer):
    """Computes and sets the pooling variables in an akida v1 layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        gap_layer (:obj:`quantizeml.Layer`): the source layer.
    """

    # In AkidaVersion.v1, biases are only applied when evaluating activations, by
    # subtracting the threshold. Since this happens after pooling, the scaling factor
    # introduced by the GAP operation must be applied to the threshold.
    spatial_size = gap_layer.input_shape[1] * gap_layer.input_shape[2]
    layer_ak.variables["threshold"] *= spatial_size


def max_pool_param_checks(max_pool_layer):
    """Check if the max_pool_layer has valid parameters for an Akida v2 conversion.
    Raise an error if not.

    Args:
        max_pool_layer (:obj:`akida.Layer`): QuantizedMaxPool2D layer to verify.

    """
    # if max_pool strides are None max_pool.strides == max_pool.pool_size => Valid assertion
    if max_pool_layer.strides is None:
        return

    pool_size = max_pool_layer.pool_size
    pool_stride = max_pool_layer.strides
    if pool_size[0] != pool_size[1]:
        raise ValueError(f"In Akida v2 {max_pool_layer.name} should have square pooling")
    if pool_stride[0] != pool_stride[1]:
        raise ValueError(f"In Akida v2 {max_pool_layer.name} pooling stride should be the same for"
                         "both dimensions")


def gap_params_checks(gap_layer):
    """Check if the gap layer has valid parameters for an Akida v1 conversion.
    Raise an error if not.

    Args:
        gap_layer (:obj:`akida.Layer`): QuantizedGlobalAveragePooling2D layer to verify.

    """

    spatial_size = gap_layer.input_shape[1] * gap_layer.input_shape[2]
    if spatial_size > 2**8:
        raise RuntimeError(f"Unsupported spatial size product ({spatial_size}). "
                           "We only accepts values smaller than 256. "
                           "Consider to replace GlobalAveragePooling2D by MaxPooling2D.")



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedConv2D to Akida.
"""

from quantizeml.layers import (QuantizedConv2D, QuantizedReLU, QuantizedGlobalAveragePooling2D,
                               QuantizedMaxPool2D)
from akida import Convolutional, Conv2D

from ..akida_versions import AkidaVersion
from .activations import set_relu_variables
from .pooling import (max_pool_param_checks, gap_params_checks,
                      set_gap_variables_v1)
from .outputs import set_output_v1_variables, set_output_v2_variables
from .padding import check_conv_and_max_pool_compatibility
from .blocks import get_block_out_quantizer
from .block_converter import BlockConverter, register_conversion_patterns
from .conv_common import (parse_conv_block_v1, get_layer_by_type,
                          parse_conv_block_v2, set_conv_variables_v1, set_conv_variables_v2)

__all__ = ["ConvBlockConverterV1", "ConvBlockConverterV2"]

_PATTERNS_V1 = [(QuantizedConv2D,), (QuantizedConv2D, QuantizedReLU),
                (QuantizedConv2D, QuantizedMaxPool2D, QuantizedReLU),
                (QuantizedConv2D, QuantizedGlobalAveragePooling2D, QuantizedReLU),
                (QuantizedConv2D, QuantizedGlobalAveragePooling2D)]

_PATTERNS_V2 = [(QuantizedConv2D,), (QuantizedConv2D, QuantizedReLU),
                (QuantizedConv2D, QuantizedMaxPool2D, QuantizedReLU),
                (QuantizedConv2D, QuantizedReLU, QuantizedGlobalAveragePooling2D),
                (QuantizedConv2D, QuantizedGlobalAveragePooling2D)]


def convert_conv_block_v1(model_ak, block):
    """Converts a convolutional block into an akida v1 Convolutional layer.

    The expected sequence is:

    - QuantizedConv2D,
    - QuantizedMaxPooling2D or QuantizedGlobalAveragePooling2D (optional),
    - QuantizedReLU (optional).

    Args:
        model_ak (:obj:`akida.Model`): the target Akida model.
        block (list(:obj:`tf.keras.Layer`)): the block layers.

    Returns:
        bool: Returns True for a successful conversion.
    """

    conv = block[0]

    # Evaluate the convolutional block layers parameters
    conv_params = parse_conv_block_v1(block)

    # Add layer to the model to build its internal variables
    conv_ak = Convolutional(**conv_params)
    model_ak.add(conv_ak)

    # Set base variables
    set_conv_variables_v1(conv_ak, conv)

    # Check if we have GAP
    pool_layer = get_layer_by_type(block, QuantizedGlobalAveragePooling2D)
    if pool_layer:
        set_gap_variables_v1(conv_ak, pool_layer)

    # Get out_quantizer of the block.
    out_quantizer = get_block_out_quantizer(block)

    if out_quantizer:
        set_output_v1_variables(conv_ak, out_quantizer)

    return True


def convert_conv_block_v2(model_ak, block):
    """Converts a convolutional block into an akida v2 Conv2D layer.

    The expected sequence is:

    - QuantizedConv2D,
    - QuantizedMaxPooling2D (optional),
    - QuantizedReLU (optional).

    or:
    - QuantizedConv2D,
    - QuantizedReLU (optional),
    - QuantizedGlobalAveragePooling2D (optional).

    Args:
        model_ak (:obj:`akida.Model`): the target Akida model.
        block (list(:obj:`tf.keras.Layer`)): the block layers.

    Returns:
        bool: Returns True for a successful conversion.
    """

    conv = block[0]

    # Evaluate the convolutional block layers parameters
    conv_params = parse_conv_block_v2(block)

    # Add layer to the model to build its internal variables
    conv_ak = Conv2D(**conv_params)
    model_ak.add(conv_ak)

    # Set base variables
    set_conv_variables_v2(conv_ak, conv)

    # Check if we have ReLU
    relu_layer = get_layer_by_type(block, QuantizedReLU)
    if relu_layer:
        set_relu_variables(conv_ak, relu_layer)

    # Get out_quantizer of the block.
    out_quantizer = get_block_out_quantizer(block)

    if out_quantizer:
        set_output_v2_variables(conv_ak, out_quantizer)

    return True


class ConvBlockConverterV1(BlockConverter):
    """Main class that should be used to check if the conv block is compatible to an Akida v1
    conversion and provides a method to convert it in an equivalent Akida Convolutional layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        self._conv_additional_checks()

    def _conv_additional_checks(self):
        conv = self._block[0]

        # The only weight bitwidth supported is [2, 4]
        weight_bits = conv.weight_quantizer.bitwidth
        if weight_bits not in [1, 2, 4]:
            raise ValueError("Convolutional layer can only handle weights with 1, 2 or 4 bits."
                             f" Received: {weight_bits}.")
        groups = getattr(conv, "groups")
        if groups != 1:
            raise ValueError("QuantizedConv2D can only be converted with groups==1."
                             f" Received groups=={groups}")
        pool_layer = get_layer_by_type(self._block, (QuantizedGlobalAveragePooling2D,
                                                     QuantizedMaxPool2D))
        if isinstance(pool_layer, QuantizedMaxPool2D):
            check_conv_and_max_pool_compatibility(conv, pool_layer)
        elif isinstance(pool_layer, QuantizedGlobalAveragePooling2D):
            gap_params_checks(pool_layer)

    def convert(self, model_ak):
        return convert_conv_block_v1(model_ak, self._block)


class ConvBlockConverterV2(BlockConverter):
    """Main class that should be used to check if the conv block is compatible to an Akida v2
    conversion and provides a method to convert it in an equivalent Akida Conv2D layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        self._conv_additional_checks()

    def _conv_additional_checks(self):
        conv = self._block[0]
        pool_layer = get_layer_by_type(self._block, QuantizedMaxPool2D)
        if pool_layer:
            check_conv_and_max_pool_compatibility(conv, pool_layer)
            max_pool_param_checks(pool_layer)

        # The only weight bitwidth supported is [4, 8]
        weight_bits = conv.weight_quantizer.bitwidth
        if weight_bits not in [4, 8]:
            raise ValueError("Conv2D layer can only handle weights with 4 or 8 bits."
                             f" Received: {weight_bits}.")

        # Make sure the Conv2D layers spatial params are square
        assert conv.kernel_size[0] == conv.kernel_size[1], ("Conv2D handle only"
                                                            "square kernels")
        assert conv.strides[0] == conv.strides[1], ("Conv2D stride should be"
                                                    "the same for both dimensions")
        groups = getattr(conv, "groups")
        if groups != 1:
            raise ValueError("QuantizedConv2D can only be converted with groups==1."
                             f" Received groups=={groups}")

    def convert(self, model_ak):
        return convert_conv_block_v2(model_ak, self._block)


# Register the valid conv block patterns for Akida v1 and v2
register_conversion_patterns(AkidaVersion.v1, _PATTERNS_V1, ConvBlockConverterV1)
register_conversion_patterns(AkidaVersion.v2, _PATTERNS_V2, ConvBlockConverterV2)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["get_akida_padding"]

import akida
from quantizeml.onnx_support.layers.compute_shapes import compute_conv_output


def compute_same_pads(input_shape, kernel_shape, strides):
    """Compute pads values.

    Args:
        input_shape (tuple of ints): the input shape.
        kernel_shape (tuple of ints): the convolutional kernel shape.
        strides (tuple of ints): the convolutional strides.

    Returns:
        tuple of ints: the pads to apply.
    """
    x, y = input_shape
    filter_x, filter_y = kernel_shape
    stride_x, stride_y = strides

    if x % stride_x == 0:
        pad_along_x = max(filter_x - stride_x, 0)
    else:
        pad_along_x = max(filter_x - (x % stride_x), 0)
    if y % stride_y == 0:
        pad_along_y = max(filter_y - stride_y, 0)
    else:
        pad_along_y = max(filter_y - (y % stride_y), 0)
    pad_y1 = pad_along_y // 2
    pad_y2 = pad_along_y - pad_y1
    pad_x1 = pad_along_x // 2
    pad_x2 = pad_along_x - pad_x1
    return [pad_x1, pad_y1, pad_x2, pad_y2]


def get_akida_padding(converter):
    """Returns the akida padding

    Args:
        converter (OnnxConverter): the converter to extract the padding.

    Returns:
        akida.Padding: the padding
    """
    # Check pads are compatible with akida.Padding.Same
    kernel_shapes = converter.weights["W"].shape[-2:]
    if any(x for x in converter.pads):
        exp_pads = compute_same_pads(converter.input_shape[:2], kernel_shapes, converter.strides)
        ak_padding = akida.Padding.Same
    else:
        exp_pads = [0] * 2 * len(kernel_shapes)
        ak_padding = akida.Padding.Valid
    # Pads in FC dimensions have to be zero
    fc_pads = converter.pads[:2] + converter.pads[4:6]
    if any(fc_pads):
        raise ValueError(f"Unrecognized {converter.pads} pads in {converter.name}.")
    # Compare if convolutional padding produces same behavior than Akida
    convert_pads = converter.pads[2:4] + converter.pads[6:]
    if convert_pads != exp_pads:
        raise ValueError(f"Expect pads {exp_pads} (found {convert_pads}) in {converter.name}.")

    # Compare if max pool padding produces same behavior than Akida
    if converter.pool_type == akida.PoolType.Max:
        # Calculate convolutional output shape
        out_shape = compute_conv_output(converter.input_shape[:2],
                                        kernel_shapes,
                                        converter.strides,
                                        convert_pads)
        if ak_padding == akida.Padding.Same:
            exp_pads = compute_same_pads(out_shape, converter.pool_size, converter.pool_strides)
        if converter.pool_pads != exp_pads:
            raise ValueError(f"Expect pads {exp_pads} (found {converter.pool_pads}) "
                             f"in {converter.name} maxpool.")

    return ak_padding
